'''
Module :samp:`twutils.QO.stationary`
====================================

This module computes stationary states of numerical atoms.
It can be used to generate state files for use with quantum optics modules.

Classes and functions
----------------------
'''

import numpy as np
import scipy as sp
from scipy.sparse.linalg import eigs
from scipy import constants as C
from scipy.optimize import newton

def DiracHypergeometricArgument(w,cylindrical,Qnuc,qorb,morb,nr,jam,jzam,Bz):
	"""Auxilliary function used to find weak field energy levels.  At present cylindrical only.
	Energy is the solution of f(w)==0.

	:parameter float w: energy of the level, depending on usage possibly guess
	:parameter boolean cylindrical: is this a cylincrical atom (currently must be true)
	:parameter float Qnuc: charge of the nucleus
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter float nr: radial quantum number
	:parameter float jam: total angular momentum
	:parameter float jzam: total angular momentum projection
	:parameter float Bz: external magnetic field (must be weak)"""
	# write the square roots so we dont get complex numbers
	# we are just trying to get the root finder to not bail
	l0 = np.sqrt(np.abs(morb**2 - w**2))
	Zw = qorb*Qnuc*w/l0
	Zm = qorb*Qnuc*morb/l0
	Omega = qorb*Bz/2
	sigma = np.sqrt(np.abs(jzam**2-(qorb*Qnuc)**2))
	beta = np.sqrt(np.abs(l0**2 - 2*jzam*Omega))
	sgn = -1.0
#	  if jzam>0.0:
#		beta = np.sqrt(np.abs(l0**2 - 2*Omega*(jzam+(sigma+Zw)/(jzam-Zm))))
#		sgn = -1.0
#	  else:
#		beta = np.sqrt(np.abs(l0**2 - 2*Omega*(jzam+(sigma-Zw)/(jzam+Zm))))
#		sgn = -1.0
	if cylindrical:
		return ((sgn+2*Zw)*l0+(1+2*sigma)*beta)/(2*beta) + nr
	else:
		return 0.0

def SchroedingerEnergyLevel(Qnuc,qorb,morb,nr,lam):
	"""Returns Schroedinger energy levels indexed by nr and lam.  The principle quantum number is nr+lam+1.

	:parameter float Qnuc: charge of the nucleus
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter int nr: radial quantum number [0,1,2,...]
	:parameter in lam: orbital angular momentum [0,1,2,...]"""
	return morb*(1.0 - (Qnuc*qorb)**2/(2.0*(lam+nr+1.0)**2))

def KGEnergyLevel(cylindrical,Qnuc,qorb,morb,nr,lam,lzam,Bz):
	"""Returns KG energy level indexed by nr,lam,lzam.
	Spherical stability of an electronic atom requires Z<=68.
	Cylindrical stability requires lam>0.

	:parameter boolean cylindrical: is this a cylincrical atom
	:parameter float Qnuc: charge of the nucleus
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter int nr: radial quantum number [0,1,2,...]
	:parameter int lam: orbital angular momentum [0,1,2,...]
	:parameter int lzam: orbital angular momentum projection [-lam,...,lam]
	:parameter float Bz: magnetic field (must be weak)"""
	if cylindrical:
		if lzam**2 - (Qnuc*qorb)**2 < 0.0:
			return 0.0
		e2i = 1.0 + (Qnuc*qorb)**2/(np.sqrt(lzam**2 - (Qnuc*qorb)**2) + nr + 0.5)**2
	else:
		if (lam+0.5)**2 - (Qnuc*qorb)**2 < 0.0:
			return 0.0
		e2i = 1.0 + (Qnuc*qorb)**2/(np.sqrt((lam+0.5)**2 - (Qnuc*qorb)**2) + nr + 0.5)**2
	return morb*np.sqrt(1.0-lzam*qorb*Bz/morb**2)/np.sqrt(e2i)

def DiracEnergyLevel(cylindrical,Qnuc,qorb,morb,nr,jam,jzam,Bz,energy_estimate=0.99):
	"""Returns Dirac energy levels indexed by nr,jam,jzam.
	Spherical stability of an electronic atom requires Z<=137.
	Cylindrical stability requires Z<=68.

	:parameter boolean cylindrical: is this a cylincrical atom
	:parameter float Qnuc: charge of the nucleus
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter int nr: radial quantum number [0,1,2,...]
	:parameter float jam: total angular momentum [0.5,1.5,2.5,...]
	:parameter float jzam: total angular momentum projection [-jam,...,jam]
	:parameter float Bz: magnetic field (must be weak)"""
	if Bz==0:
		x2 = (Qnuc*qorb)**2
		if cylindrical:
			if jzam**2 - x2 < 0.0:
				return 0.0
			e2i = 1.0 + x2/(np.sqrt(jzam**2-x2) + nr)**2
		else:
			if (jam+0.5)**2 - x2 < 0.0:
				return 0.0
			e2i = 1.0 + x2/(np.sqrt((jam+0.5)**2 - x2) + nr)**2
		return morb/np.sqrt(e2i)
	else:
		return newton(DiracHypergeometricArgument,energy_estimate,args=(cylindrical,Qnuc,qorb,morb,nr,jam,jzam,Bz))

def nu_to_au(energy):
	'''take energy from natural to atomic units'''
	return (energy - 1.0)*C.m_e*C.c**2/(2.0*C.Rydberg*C.c*C.h)

def nu_to_eV(energy):
	'''take energy from natural units to eV'''
	return (energy - 1.0)*C.m_e*C.c**2/C.e

def nu_to_keV(energy):
	'''take energy from natural units to keV'''
	return 1e-3*(energy - 1.0)*C.m_e*C.c**2/C.e

def nu_to_MeV(energy):
	'''take energy from natural units to MeV'''
	return 1e-6*(energy - 1.0)*C.m_e*C.c**2/C.e

def SoftCorePotential(Qnuc,r0,num_pts,dr):
	"""Create scalar potential on a radial grid

	:parameter float Qnuc: charge of the nucleus
	:parameter float r0: soft core radius
	:parameter int num_pts: number of interior grid points, total is num_pts+2
	:parameter float dr: radial grid spacing
	:returns: an array of radial points and an array of potentials, including ghost cells
	:returnType: tuple with the two arrays (grid,phi)"""
	grid = np.linspace(-0.5*dr,num_pts*dr + 0.5*dr,num_pts+2)
	phi = Qnuc/np.sqrt(r0**2 + grid**2)
	return grid , phi

def KGMatrix(cylindrical,grid,phi,qorb,morb,lam,lzam,Bz):
	"""Returns matrix of the discretized time-independent KG operator.
	This uses the leapfrog friendly Hamiltonian representation.
	For N grid points we obtain a 2Nx2N matrix corresponding to 2 components.
	The second component is :math:`\chi = (\omega-q\phi)\psi/m`

	:parameter boolean cylindrical: is this a cylincrical atom
	:parameter array grid: grid array, can get from SoftCorePotential function
	:parameter array phi: potential array, can get from SoftCorePotential function
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter int lam: orbital angular momentum [0,1,2,...], ignored in cylindrical case
	:parameter int lzam: orbital angular momentum projection [-lam,...,lam], or any integer in cylindrical case
	:parameter float Bz: magnetic field (must be weak in spherical case)"""
	N = grid.shape[0]
	num_pts = N-2
	dr = grid[1]-grid[0]
	r1_array = grid[1:N-1] - 0.5*dr
	r2_array = grid[1:N-1]
	r3_array = grid[1:N-1] + 0.5*dr
	phi_array = phi[1:N-1]
	if cylindrical:
		T1 = -(1.0/dr**2)*r1_array[1:]/r2_array[1:]
		T2 = morb*morb + lzam**2/r2_array**2 + (1.0/dr**2)*(r1_array + r3_array)/r2_array
		T2 -= lzam*qorb*Bz - (0.5*qorb*Bz*r2_array)**2
		T3 = -(1.0/dr**2)*r3_array[:num_pts-1]/r2_array[:num_pts-1]
	else:
		T1 = -(1.0/dr)*3.0*r1_array[1:]**2/(r3_array[1:]**3-r1_array[1:]**3)
		T2 = morb*morb + lam*(lam+1)/r2_array**2 + (1.0/dr)*3.0*(r1_array**2 + r3_array**2)/(r3_array**3-r1_array**3)
		T2 -= lzam*qorb*Bz # have to neglect B^2 or we would need to solve in r-theta plane
		T3 = -(1.0/dr)*3.0*r3_array[:num_pts-1]**2/(r3_array[:num_pts-1]**3-r1_array[:num_pts-1]**3)
	# may have to use "spdiags" with older scipy
	mA = sp.sparse.diags([qorb*phi_array],[0])
	mB = sp.sparse.diags([np.ones(num_pts)*morb],[0])
	mC = sp.sparse.diags([T1/morb,T2/morb,T3/morb],[-1,0,1])
	mD = sp.sparse.diags([qorb*phi_array],[0])
	return sp.sparse.bmat([[mA,mB],[mC,mD]])

def DiracMatrix(cylindrical,grid,phi,qorb,morb,jam,lam,jzam,Bz):
	"""Returns matrix of the discretized time-independent Dirac operator.
	For N grid points we obtain a 2Nx2N matrix corresponding to 2 components.
	The two components are :math:`f` and :math:`g` which in turn form the bispinor.
	See, e.g., Landau and Lifshitz QED section 35.

	:parameter boolean cylindrical: is this a cylincrical atom
	:parameter array grid: grid array, can get from SoftCorePotential function
	:parameter array phi: potential array, can get from SoftCorePotential function
	:parameter float qorb: charge of the orbiting particle
	:parameter float morb: mass of the orbiting particle
	:parameter float jam: total angular momentum [0.5,1.5,...], ignored in cylindrical case
	:parameter int lam: parity number, must be jam :math:`\pm 1/2`, or in cylindrical case, jzam :math:`\pm 1/2`
	:parameter float jzam: total angular momentum projection [-jam,...,jam], or in cylindrical case, any half integer
	:parameter float Bz: magnetic field (must be weak in spherical case)"""
	N = grid.shape[0]
	num_pts = N-2
	# uniform grid is assumed
	# may have to use "spdiags" with older scipy
	dr = grid[1:N-1]-grid[0:N-2]
	rc = grid[1:N-1]
	phi_array = phi[1:N-1]
	if cylindrical:
		kappa = 2*(lam-jzam)*(jzam - 0.5*qorb*Bz*rc*rc)
		D1 = np.complex(-0.5)/dr[1:]
		D2f = np.complex(0.5)/rc
		D2g = np.complex(0.5)/rc
		D3 = np.complex(0.5)/dr[:num_pts-1]
		# Boundary condition based on parity number
		D2f[0] += D1[0]*(-1)**lam
		D2g[0] -= D1[0]*(-1)**lam
		mA = sp.sparse.diags([np.complex(qorb)*phi_array+morb],[0])
		mB = sp.sparse.diags([-1j*D1,-1j*(D2g-kappa/rc),-1j*D3],[-1,0,1])
		mC = sp.sparse.diags([-1j*D1,-1j*(D2f+kappa/rc),-1j*D3],[-1,0,1])
		mD = sp.sparse.diags([np.complex(qorb)*phi_array-morb],[0])
	else:
		if Bz!=0.0:
			exit(1)
		# spherical case ignores B-field
		kappa = 2*(lam-jam)*(jam+0.5)
		D1 = -0.5/dr[1:]
		D2f = 1.0/rc
		D2g = 1.0/rc
		D3 = 0.5/dr[:num_pts-1]
		# Boundary condition based on parity number
		# But what are the parities of f & g ??
		D2f[0] += D1[0]*(-1)**lam
		D2g[0] -= D1[0]*(-1)**lam
		mA = sp.sparse.diags([np.complex(qorb)*phi_array+morb],[0])
		mB = sp.sparse.diags([-D1,-D2g+kappa/rc,-D3],[-1,0,1])
		mC = sp.sparse.diags([D1,D2f+kappa/rc,D3],[-1,0,1])
		mD = sp.sparse.diags([np.complex(qorb)*phi_array-morb],[0])
	return sp.sparse.bmat([[mA,mB],[mC,mD]])

class StationaryStateTool:
	'''Class for managing stationary states, and creating state files'''

	def __init__(self,cylindrical,grid,phi,nr,jam,lam,jzam):
		'''Create a stationary state tool.

		:parameter bool cylindrical: is this a cylindrical atom
		:parameter array grid: radial grid, can be output from SoftCorePotential()
		:parameter array phi: potential on the radial grid, can be output from SoftCorePotential()
		:parameter int nr: radial quantum number [0,1,2,...] (cannot be 0 in certain cases)
		:parameter float jam: see lower level functions for possible values
		:parameter int lam: see lower level functions for possible values
		:parameter float jzam: see lower level functions for possible values.  This triggers Dirac or KG treatment depending on whether integer or half-integer.'''
		self.cylindrical = cylindrical
		self.grid = grid
		self.phi = phi
		self.nr = nr
		self.jam = jam
		self.lam = lam
		self.jzam = jzam
		self.qorb = -np.sqrt(C.alpha)
		self.morb = 1.0
		self.Bz = 0.0

	def SetSpecialConditions(self,qorb,morb,Bz):
		self.qorb = qorb
		self.morb = morb
		self.Bz = Bz

	def RescalePhi(self,scale_factor):
		self.phi *= scale_factor

	def CheckQuantumNumbers(self):
		nr = self.nr
		jam = self.jam
		lam = self.lam
		jzam = self.jzam
		if np.abs(np.modf(jzam)[0])!=0.5 and np.abs(np.modf(jzam)[0])!=0.0:
			return 'jzam is neither integer nor half-integer'

		# Spin 1/2 Case
		if np.abs(np.modf(jzam)[0])==0.5:
			if self.cylindrical:
				if np.abs(jzam-lam)!=0.5:
					return 'abs(jzam-lam) is not 1/2.'
				if nr==0.0 and (lam-jzam)*jzam>0.0:
					return 'nr cannot be 0 with requested angular mode.'
			else:
				if np.abs(np.modf(jam)[0])!=0.5 or jam<0.5:
					return 'jam is not a positive half-integer.'
				if np.abs(jam-lam)!=0.5:
					return 'abs(jam-lam) is not 1/2.'
				if nr==0.0 and (lam-jam)*(jam+0.5)>0.0:
					return 'nr cannot be 0 with requested angular mode.'
				if jzam<-jam or jzam>jam:
					return 'violation of -jam<=jzam<=jam'

		# Spin 0 Case
		if np.abs(np.modf(jzam)[0])==0.0:
			if not self.cylindrical:
				if np.abs(np.modf(lam)[0])!=0.0 or lam<0.0:
					return 'lam is not a positive integer or zero.'
				if jzam<-lam or jzam>lam:
					return 'violation of -lam<=jzam<=lam'

		return 'OK'

	def spin0(self):
		return np.abs(np.modf(self.jzam)[0])==0.0

	def GetEnergy(self,energy_guess,scale_factor):
		nr_max = np.int(self.nr + 1)
		# Find requested state; sorting energies identifies nr
		if self.spin0():
			mat = KGMatrix(self.cylindrical,self.grid,scale_factor*self.phi,self.qorb,self.morb,self.lam,self.jzam,self.Bz)
		else:
			mat = DiracMatrix(self.cylindrical,self.grid,scale_factor*self.phi,self.qorb,self.morb,self.jam,self.lam,self.jzam,self.Bz)
		vals,vecs = eigs(mat.tocsc(),k=nr_max,sigma=energy_guess,which='LM')
		sorted_indices = np.argsort(vals)
		idx = sorted_indices[np.int(self.nr)]
		En = np.real(vals[idx])
		return En

	def GetBaseComponents(self,energy_guess):
		'''Gets 2 components used to form the wavefunction, call them f and g.
		For KG returns a Hamiltonian decomposition, with f the usual scalar, and g an auxiliary function.
		The Feshbach-Villars decomposition is then :math:`(f+g)/\sqrt{2}` , :math:`(f-g)/\sqrt{2}`.
		For Dirac returns f,g from L&L, which may be used to form the bispinor.

		:parameter float energy_guess: initial guess of the energy
		:returns: En,f,g,sorted (En is the selected eigenvalue, sorted are the first few eigenvalues)'''

		num_pts = self.grid.shape[0]-2
		nr_max = np.int(self.nr + 4)

		# Find requested state; sorting energies identifies nr

		if self.spin0():
			mat = KGMatrix(self.cylindrical,self.grid,self.phi,self.qorb,self.morb,self.lam,self.jzam,self.Bz)
		else:
			mat = DiracMatrix(self.cylindrical,self.grid,self.phi,self.qorb,self.morb,self.jam,self.lam,self.jzam,self.Bz)
		vals,vecs = eigs(mat.tocsc(),k=nr_max,sigma=energy_guess,which='LM')

		sorted_indices = np.argsort(vals)
		idx = sorted_indices[np.int(self.nr)]
		f = vecs[:num_pts,idx]
		g = vecs[num_pts:,idx]
		En = np.real(vals[idx])
		sorted = np.sort(np.real(vals))
		return En,f,g,sorted

	def GetWavefunction(self,f,g):
		'''Return 4 components (0,1,2,3) derived from the base components f and g.
		For KG only components 1 and 2 are relevant, components 0 and 3 may be ignored.
		For KG we return the simple decomposition with psi1 the usual scalar wavefunction.
		The Feshbach-Villars decomposition is then :math:`(\psi_1+\psi_2)/\sqrt{2}` , :math:`(\psi_1-\psi_2)/\sqrt{2}`.
		For Dirac the 4 components are the radial part of the bispinor.'''
		nr = self.nr
		jam = self.jam
		lam = self.lam
		jzam = self.jzam
		num_pts = self.grid.shape[0]-2
		dr = self.grid[1]-self.grid[0] # assume uniform grid
		if self.cylindrical:
			jacobian = 2*np.pi*self.grid[1:num_pts+1]
		else:
			jacobian = self.grid[1:num_pts+1]**2
		if self.spin0():
			psi0 = np.zeros(num_pts)
			psi1 = f
			psi2 = g
			psi3 = np.zeros(num_pts)
			rho = 0.5*(np.abs(f+g)**2 - np.abs(f-g)**2)
		else:
			if self.cylindrical:
				if jzam>lam:
					psi0 = f
					psi1 = np.zeros(num_pts)
					psi2 = np.zeros(num_pts)
					psi3 = g
				else:
					psi0 = np.zeros(num_pts)
					psi1 = f
					psi2 = g
					psi3 = np.zeros(num_pts)
			else:
				psi0 = f
				psi1 = f
				psi2 = (-1.0)**((1+2*lam-2*jam)/2)*g
				psi3 = (-1.0)**((1+2*lam-2*jam)/2)*g
			rho = np.abs(psi0)**2 + np.abs(psi1)**2 + np.abs(psi2)**2 + np.abs(psi3)**2

		CN = np.sqrt(1/np.sum(dr*rho*jacobian))
		psi0 = CN*psi0
		psi1 = CN*psi1
		psi2 = CN*psi2
		psi3 = CN*psi3

		return psi0,psi1,psi2,psi3

	def StandardFilename(self,Z):
		'''Return a filename appropriate for the current quantum state

		:parameter float Z: atomic number, purely for labeling
		:returns: string with the filename'''
		file_name = 'z'+str(np.int(Z))
		if self.cylindrical:
			file_name += '_c'
			if self.spin0():
				angular_num = self.jzam
			else:
				angular_num = self.lam
		else:
			file_name += '_s'
			angular_num = self.lam
		file_name += str(np.int(self.nr))
		file_name += str(np.int(angular_num))
		if not self.spin0():
			if self.cylindrical:
				if self.jzam>self.lam:
					file_name += '+'
				else:
					file_name += '-'
			else:
				if self.jam>self.lam:
					file_name += '+'
				else:
					file_name += '-'
		file_name += '.txt'
		return file_name

	def WriteFile(self,file_name,psi0,psi1,psi2,psi3,En,softCoreRadius,Qnuc):
		'''Write the state file given the current quantum state'''
		num_pts = self.grid.shape[0]-2
		dr = self.grid[1]-self.grid[0] # assume uniform grid
		outFile = open(file_name,'w')
		outFile.write('energy = ')
		outFile.write(str(En))
		outFile.write('\n')
		outFile.write('pts = ')
		outFile.write(str(num_pts))
		outFile.write('\n')
		if self.spin0():
			outFile.write('components = 2')
		else:
			outFile.write('components = 4')
		outFile.write('\n')
		outFile.write('cell_width = ')
		outFile.write(str(dr))
		outFile.write('\n')
		outFile.write('soft_core_radius = ')
		outFile.write(str(softCoreRadius))
		outFile.write('\n')
		outFile.write('nuclear_charge = ')
		outFile.write(str(Qnuc))
		outFile.write('\n')
		outFile.write('nr_j_l_m = ')
		outFile.write(str(self.nr)+' ')
		outFile.write(str(self.jam)+' ')
		outFile.write(str(self.lam)+' ')
		outFile.write(str(self.jzam)+ ' ')
		outFile.write('\n')
		outFile.write('Bz = ')
		outFile.write(str(self.Bz))
		outFile.write('\n')
		if self.cylindrical:
			outFile.write('cylindrical = true')
		else:
			outFile.write('cylindrical = false')
		outFile.write('\n\n')
		outFile.write('start_data\n')
		if self.spin0():
			np.column_stack((np.real(psi1),np.imag(psi1))).tofile(outFile,'\n')
			outFile.write('\n')
			np.column_stack((np.real(psi2),np.imag(psi2))).tofile(outFile,'\n')
		else:
			np.column_stack((np.real(psi0),np.imag(psi0))).tofile(outFile,'\n')
			outFile.write('\n')
			np.column_stack((np.real(psi1),np.imag(psi1))).tofile(outFile,'\n')
			outFile.write('\n')
			np.column_stack((np.real(psi2),np.imag(psi2))).tofile(outFile,'\n')
			outFile.write('\n')
			np.column_stack((np.real(psi3),np.imag(psi3))).tofile(outFile,'\n')
		outFile.close()



import sys
import pathlib
import glob
import json
import warnings
import pkg_resources
import numpy as np
import h5py
from scipy import constants as C

def print_usage():
    print('Usage: os2tw [--help,-h] [--version,-v] <cmd> <component>[,<modifier>]')
    print('             [n1=<density>] [dt=<frame_time>] [frame=<::>] [nfac=<factor>] [root=<MS>]')
    print('Extracts data from OSIRIS directory tree, consolidates into npy file, and creates metadata file.')
    print('1D coordinate label mapping: 1->z')
    print('2D coordinate label mapping: 1->z, 2->x')
    print('3D coordinate label mapping: 1->z, 2->y, 3->x')
    print('-------------Examples----------------')
    print('Field example: os-convert.py init e1')
    print('Charge example: os-convert.py init charge,electrons')
    print('Phase space example: os-convert.py init p1x1,electrons')
    print('-------------Arguments----------------')
    print('cmd = init or append, if init restart the metadata, otherwise append to it.')
    print('component = e1,e2,e3,b1,b2,b3,etc..')
    print('modifier = used to narrow the file search, typically a species name, but could include other things.')
    print('n1 = a normalizing density in cgs units, needed to support TW unit conversions.')
    print('dt = time between frames in plasma units.')
    print('frame = python style range indicating frames to keep, e.g., 2:11:4, etc.')
    print('nfac = OSIRIS frame skipping multiplication factor')
    print('root = OSIRIS root data directory relative to working directory')

def print_version():
    print('os2tw is provided by twutils, version '+pkg_resources.get_distribution('twutils').version)

def str_to_range(rng_str,num):
    '''Convert a numpy-style slice string to a range'''
    rng = rng_str.split(':')
    tup = ()
    for i,el in enumerate(rng):
        if el=='' and i==0:
        	el = '0'
        if el=='' and i==1:
        	el = str(num)
        if el=='' and i==2:
        	el = '1'
        tup += (int(el),)
    return range(tup[0],tup[1],tup[2])

def get_data_ext(ax):
    '''Convert OSIRIS bounds data, involving some re-arrangement and padding.

    :param dict ax: result of AXIS key in OSIRIS HDF5 file'''
    data_ext = []
    data_ext += [ax['AXIS1'][0],ax['AXIS1'][1]]
    try:
        data_ext += [ax['AXIS2'][0],ax['AXIS2'][1]]
        try:
            data_ext += [ax['AXIS3'][0],ax['AXIS3'][1]]
            data_ext = data_ext[4:6] + data_ext[2:4] + data_ext[:2]
        except LookupError:
            data_ext = data_ext[2:4] + [-0.5,0.5] + data_ext[:2]
    except LookupError:
        data_ext = [-0.5,0.5,-0.5,0.5] + data_ext
    return data_ext

def main_metadata(name,meta):
    meta[name] = {}
    meta[name]['native units'] = 'plasma'
    meta[name]['grid'] = name + '_grid_warp.txt'
    meta[name]['axes'] = {}
    return meta

def axis_metadata(ax,component,n1_mks):
    wp = np.sqrt(n1_mks*C.e**2/C.epsilon_0/C.m_e)
    xa = (C.c/wp) / (C.hbar/C.m_e/C.c/C.alpha)
    ta = (1/wp) / (C.hbar/C.m_e/C.c**2/C.alpha**2)
    xn = (C.c/wp) / (C.hbar/C.m_e/C.c)
    tn = (1/wp) / (C.hbar/C.m_e/C.c**2)
    emks = C.m_e*C.c*wp/C.e
    ecgs = emks / (1e-4*C.c)
    ea = emks / (C.m_e**2*C.c**3*C.alpha**3 / C.hbar / C.e)
    en = emks / (C.m_e**2*C.c**3 / C.hbar / C.e)
    units = ['mks','cgs','plasma','atomic','natural']
    label = ['$t$','$x$','$y$','$z$','$'+component+'$']
    fields = ['e1','e2','e3','b1','b2','b3']
    spectra = ['gamma']
    phase_space_axes = ['x1','x2','x3','p1','p2','p3']
    phase_spaces = []
    for i in range(6):
        for j in range(6):
            if i!=j:
                phase_spaces += [phase_space_axes[i]+phase_space_axes[j]]
    l = {}
    l['none'] = {'mks':'AU' , 'cgs':'AU' , 'plasma':'', 'atomic':'AU', 'natural':'AU'}
    l['t'] = {'mks':'s' , 'cgs':'s' , 'plasma':r'$\omega$', 'atomic':'a.u.', 'natural':'n.u.'}
    l['x'] = {'mks':'m' , 'cgs':'cm' , 'plasma':r'$\omega/c$', 'atomic':'a.u.', 'natural':'n.u.'}
    l['p'] = {'mks':'kg$\cdot$m$/$s' , 'cgs':'g$\cdot$cm$/$s' , 'plasma':r'$/mc$', 'atomic':'a.u.', 'natural':'n.u.'}
    l['e'] = {'mks':'V$/$m' , 'cgs':'SV$/$cm' , 'plasma':r'$e/mc\omega$', 'atomic':'a.u.', 'natural':'n.u.'}
    l['b'] = {'mks':'T' , 'cgs':'G' , 'plasma':r'$e/mc\omega$', 'atomic':'a.u.', 'natural':'n.u.'}
    m = {}
    m['none'] = {'mks':1.0 , 'cgs':1.0 , 'plasma':1.0, 'atomic':1.0, 'natural':1.0}
    m['t'] = {'mks':1/wp , 'cgs':1/wp , 'plasma':1.0, 'atomic':xa, 'natural':xn}
    m['x'] = {'mks':C.c/wp , 'cgs':100*C.c/wp , 'plasma':1.0, 'atomic':xa, 'natural':xn}
    m['p'] = {'mks':C.m_e*C.c , 'cgs':1e5*C.m_e*C.c , 'plasma':1.0, 'atomic':C.m_e*C.c*C.alpha, 'natural':C.m_e*C.c}
    m['e'] = {'mks':emks , 'cgs':ecgs , 'plasma':1.0, 'atomic':ea, 'natural':en}
    m['b'] = {'mks':emks/C.c , 'cgs':ecgs , 'plasma':1.0, 'atomic':ea, 'natural':en}
    mult = [[]]*5
    ulab = [[]]*5
    if component not in fields+phase_spaces+spectra:
        warnings.warn('Requested component not known, units will be arbitrary.')
        for i,s in enumerate(units):
            mult[i] = [m['none'][s]]*5
            ulab[i] = [l['none'][s]]*5
    if component in fields:
        label[4] = '$'+component[0].upper()+'_'+component[1]+'$'
        for i,s in enumerate(units):
            mult[i] = [m['t'][s]] + [m['x'][s]]*3 + [m[component[0]][s]]
            ulab[i] = [l['t'][s]] + [l['x'][s]]*3 + [l[component[0]][s]]
    if component in phase_spaces:
        h = component[0]+'_'+component[1]
        v = component[2]+'_'+component[3]
        label[1] = '$'+h+'$'
        label[3] = '$'+v+'$'
        label[4] = '$f('+h+','+v+')$'
        for i,s in enumerate(units):
            mult[i] = [m['t'][s],m[component[0]][s],1.0,m[component[2]][s],1.0]
            ulab[i] = [l['t'][s],l[component[0]][s],l['none'][s],l[component[2]][s],l['none'][s]]
    if component in spectra:
        label[3] = r'$\gamma$'
        label[4] = r'$f(t,\gamma)$'
        for i,s in enumerate(units):
            mult[i] = [m['t'][s],1.0,1.0,1.0,1.0]
            ulab[i] = [l['t'][s],'','','','']
    a = {}
    a['label'] = label[ax]
    for i in range(len(units)):
        a[units[i]+' label'] = ulab[i][ax]
        a[units[i]+' multiplier'] = mult[i][ax]
    return a

def grid_string(frame,dt,bounds,shp):
    s = 't = ' + str(frame*dt) + '\n'
    pts = lambda x0,x1,n : np.linspace(x0+0.5*(x1-x0)/n , x1-0.5*(x1-x0)/n , n)
    for ax in range(1,4):
        s += 'axis' + str(ax) + ' ='
        for val in pts(bounds[int(2*ax-2)],bounds[int(2*ax-1)],shp[ax-1]):
            s += ' ' + str(val)
        s += '\n'
    return s

def main():
    # Process command line
    # Required arguments
    if len(sys.argv)<2 or '--help' in sys.argv or '-h' in sys.argv:
        print_usage()
        exit(0)
    if '--version' in sys.argv or '-v' in sys.argv:
        print_version()
        exit(0)
    cmd = sys.argv[1]
    if cmd not in ['init','append']:
        raise ValueError('First argument <'+cmd+'> not recognized.')
        print('ERROR: not a valid command.')
        exit(1)
    id = sys.argv[2].split(',')
    component = id[0]
    if len(id)>1:
        modifier = id[-1]
    else:
        modifier = ''
    # Defaults for optional arguments
    range_str = '::'
    root_dir = 'MS'
    n1_mks = 1e19
    dt = 1.0
    nfac = 1
    keys_found = []
    # Process optional arguments
    for keyval in sys.argv[3:]:
        key = keyval.split('=')[0]
        val = keyval.split('=')[1]
        keys_found += [key]
        if key not in ['frame','root','n1','dt','nfac']:
            raise KeyError('Command line argument <'+key+'> not recognized.')
        if key=='frame':
            range_str = val
        if key=='root':
            root_dir = val
        if key=='n1':
            n1_mks = 1e6*np.float(val)
        if key=='dt':
            dt = np.float(val)
        if key=='nfac':
            nfac = int(val)
    if 'n1' not in keys_found:
        warnings.warn('No density given, physical units will be arbitrary.')
    if 'dt' not in keys_found:
        warnings.warn('No time step given, time axis will be arbitrary.')

    # Search for the desired OSIRIS data file
    if modifier=='':
        glob_str = root_dir + '/**/' + component + '*[0-9][0-9][0-9][0-9][0-9][0-9].h5'
    else:
        glob_str = root_dir + '/**/' + component + '*' + modifier + '*[0-9][0-9][0-9][0-9][0-9][0-9].h5'
    file_list = glob.glob(glob_str,recursive=True)
    num_frames = len(file_list)
    if num_frames==0:
        print('ERROR: no files matched the requested pattern.')
        exit(1)
    prefix_list = list(set([s[:-9] for s in file_list]))
    if len(prefix_list)>1:
        print('ERROR: found more than one match, use modifier to narrow search.')
        for i,res in enumerate(prefix_list):
            print('    Match '+str(i)+': '+res)
        exit(1)
    full_path_prefix = prefix_list[0]
    prefix = pathlib.Path(full_path_prefix).name
    print('Found {:04d} h5 files with prefix {}.'.format(num_frames,prefix))

    # Set up the frames to grab
    frame_list = []
    for frame in str_to_range(range_str,num_frames):
        frame_list += [frame]

    # Get the axis and dimension information based on first frame

    try:
        f = h5py.File(file_list[0],'r')
    except (FileNotFoundError,PermissionError):
        print('Could not open '+file_list[0])
        exit(1)

    print('Found HDF5 groups',list(f.keys()))
    ax = f['AXIS']
    data_key = None
    for k in f.keys():
        if component in k:
            data_key = k
    if data_key==None:
        print('ERROR: could not match to an HDF5 key.')
        exit(1)

    dset = f[data_key]
    print('Dimensions',dset.shape)
    print('Initial data bounds',get_data_ext(ax))
    print('Initial value range',np.min(dset),np.max(dset))

    # name of the file to create
    new_file = prefix[:-1]+'.npy'

    # TW shapes are always 4 dimensional, so add necessary axes
    if len(dset.shape)==1:
    	new_shape = (1,1) + dset.shape
    if len(dset.shape)==2:
    	new_shape = (dset.shape[0],1,dset.shape[1])
    if len(dset.shape)==3:
    	new_shape = dset.shape
    full_array = np.zeros((len(frame_list),)+new_shape)
    grid_str = ''
    count = 0
    for n in frame_list:
        fullpath = full_path_prefix + '{:06d}.h5'.format(n*nfac)
        try:
            f = h5py.File(fullpath,'r')
        except (FileNotFoundError,PermissionError):
            print('Could not open HDF5 file <' + fullpath + '>.')
            exit(1)
        full_array[count,...] = np.array(f[data_key]).reshape(new_shape)
        grid_str += grid_string(n*nfac,dt,get_data_ext(f['AXIS']),new_shape)
        count += 1
        print('.',end='',flush=True)

    # Save the data into the .npy file
    np.save(new_file,full_array)
    print()

    # create the metadata

    if cmd=='init':
        meta = {}
    else:
        with open('tw_metadata.json','r') as f:
            meta = json.load(f)
    meta = main_metadata(new_file,meta)
    for ax in range(5):
        meta[new_file]['axes'][str(ax)] = axis_metadata(ax,component,n1_mks)
    with open('tw_metadata.json','w') as f:
        json.dump(meta,f,indent=4)

    with open(new_file+'_grid_warp.txt','w') as f:
        f.write(grid_str)



import os
import shutil
import sys
import glob
import subprocess
import re
import json
import numpy as np

# Scripted arbitrary dimensional parameter scans.
# Works either as parallel-parallel (typically HPC), or parallel-serial (typically workstation).
# Premise is to change the values of #define constants for each run.
# The values of the constants are determined by a scan file, which is a JSON serialized
# dictionary, whose keys are the identifiers, and whose values are nested lists of floats.

def print_usage():
    print('Usage: twscan <scan_file> [optional arguments...]')
    print('<scan_file> : JSON serialized dictionary of parameters and values')
    print('-n : number of MPI tasks (parallel-serial only)')
    print('-c : number of OpenMP threads per MPI task (parallel-serial only)')
    print('--dry-run : only create files, do not submit jobs')
    print('--input-file,-i : name of input file')
    print('--launch,-l: MPI launcher, if needed')
    print('--submit: job submission command, if needed')
    print('--script: job submission script, if needed')
    print('--help,-h : this message')

def decode_job(n,dims):
    '''Return index space coordinate for job n as a list'''
    stride = [1]
    for i in range(1,len(dims)):
        stride += [stride[i-1]*dims[i-1]]
    idx = list(range(len(dims)))
    idx[len(dims)-1] = int(n / stride[len(dims)-1])
    offset = 0
    for i in range(len(dims)-2,-1,-1):
        offset += idx[i+1]*stride[i+1]
        idx[i] = int((n-offset)/stride[i])
    return idx

def main():
    exec_name = 'tw3d'
    input_name = 'stdin'
    script_name = ''
    submit = ''
    launch = ''
    regex_dict = {   '' : '',
                    'qsub' : r'^#PBS\s+-N\s+',
                    'sbatch' : r'^#SBATCH\s+(--job-name|-J)\s*=\s*' }
    name_dict = {   '' : '',
                    'qsub' : r'#PBS -N ',
                    'sbatch' : r'#SBATCH --job-name=' }

    # Process command line arguments

    if len(sys.argv)<2 or '--help' in sys.argv or '-h' in sys.argv:
        print_usage()
        exit(0)

    with open(sys.argv[1],'r') as f:
        substitutions = json.load(f)
    keys = list(substitutions.keys())
    dims = np.array(substitutions[keys[0]]).shape
    njobs = 1
    for s in dims:
        njobs  = njobs * s
    print('Dimensions of scan =',dims)
    print('Parameters to vary =',keys)

    dry_run = False
    prev = ''
    mpi_tasks = 1
    omp_threads = 1

    for arg in sys.argv[2:]:
        if prev=='' and arg not in ['--dry-run','-h','--help','-n','-c','-i','--input-file','-l','--launch','--submit','--script']:
            raise ValueError('Argument '+arg+' not recognized.')
        if prev=='-n':
            mpi_tasks = int(arg)
        if prev=='-c':
            omp_threads = int(arg)
        if prev=='--input-file' or prev=='-i':
            input_name = arg
        if prev=='--launch' or prev=='-l':
            launch = arg
        if prev=='--submit':
            submit = arg
        if prev=='--script':
            script_name = arg
        if arg=='--dry-run':
            dry_run = True
        if arg in ['-n','-c','-i','--input-file','-l','--launch','--submit','--script']:
            prev = arg
        else:
            prev = ''
    name_re = regex_dict[submit]
    name_cmd = name_dict[submit]

    # Look for needed files

    if submit!='':
        check = glob.glob(exec_name)
        if len(check)==0:
            raise OSError('Required executable "'+exec_name+'" is not in the working directory.')

        check = glob.glob(script_name)
        if len(check)==0:
            raise OSError('Required batch script "'+script_name+'" is not in the working directory.')

        with open(script_name,'r') as f:
            data = f.read()
            if data.find('\ncd ')>=0:
                raise Exception('The batch script changes directories, this is not allowed.')
            if data.find('./'+exec_name)==-1:
                raise Exception('Could not find ./'+exec_name+' in batch script.')
            if data.find('\n'+launch)==-1:
                raise Exception('Could not find [LF]'+launch+' in batch script.')

    check = glob.glob(input_name)
    if len(check)==0:
        raise OSError('Required input file "'+input_name+'" is not in the working directory.')

    # Check existing directories

    rundirs = glob.glob('run*')
    if len(rundirs)>0:
        ans = ''
        while ans!='y' and ans!='n':
            ans = input('Found some files or directories in the form run*, OK to clean (y/n) ?')
        if ans=='n':
            print('STOPPED. Please run script in a directory where there are no important files of the form run*.')
            exit(1)
        for rundir in rundirs:
            shutil.rmtree(rundir)

    # Carry out the scan

    root_dir = os.getcwd()

    for n in range(njobs):
        idx = decode_job(n,dims)
        job_str = ''
        for i in idx:
            job_str += '_{:04d}'.format(i)
        rundir = root_dir+'/run'+job_str
        os.mkdir(rundir)
        shutil.copy(exec_name,rundir)

        # Modify the input file
        with open(root_dir+'/'+input_name,'r') as f:
            data = f.read()
            for key in substitutions:
                esc_key = re.sub(r'\$',r'\$',key)
                val = np.array(substitutions[key])[tuple(idx)]
                macro_def = r'^\s*#define\s+'+esc_key+r'\s+.*\S\s*$'
                if re.search(macro_def,data,flags=re.MULTILINE)==None:
                    raise Exception('no match for parameter ' + key)
                data = re.sub(macro_def,'#define '+key+' '+str(val),data,flags=re.MULTILINE)
            with open(rundir+'/'+input_name,'w') as g:
                g.write(data)

        if submit=='':
            # This is a serial batch
            if dry_run:
                print('dry run: setup',rundir,'(no execution)')
            else:
                print('Executing job',job_str)
                print('-------------------------------\n')
                os.chdir(rundir)
                subprocess.run([exec_name,'-n',str(mpi_tasks),'-c',str(omp_threads),'--no-interactive','--input-file',input_name])
                os.chdir(root_dir)
        else:
            # This is a parallel batch
            # Modify the batch script
            script_path_out = rundir+'/'+script_name
            with open(root_dir+'/'+script_name,'r') as f:
                data = f.read()
                # Add the change of directory
                new_script = re.sub(r'^\s*'+launch,'\ncd '+rundir+'\n'+launch,data,flags=re.MULTILINE)
                # Change the job name
                new_script = re.sub(name_re+r'\w+',name_cmd+'tw'+job_str,new_script,flags=re.MULTILINE)
                with open(script_path_out,'w') as g:
                    g.write(new_script)
            # Submit the job
            if dry_run:
                print('dry run: setup',rundir,'(no job submission)')
            else:
                print('Submitting '+script_path_out)
                subprocess.run([submit,script_path_out])



'''This program is used to interactively install turboWAVE core once
the tools have already been installed.  This can be run in two modes:
1. graphical interface based on the tkinter module (default).
2. textual interface based on the curses module (if --terminal argument is given).
Both interfaces are built on the 3 objects config, tasks, cmd.
The config object displays the configuration.
The tasks object displays the list of tasks.
The cmd object accepts inputs, and displays messages.'''

import time
import os
import sys
import shutil
import pathlib
import glob
import subprocess
import re
import copy
import pkg_resources
try:
    import curses
    has_curses = True
except ModuleNotFoundError:
    has_curses = False
try:
    import tkinter as tk
    from tkinter import ttk
    import tkinter.messagebox
    import tkinter.filedialog
    has_tk = True
except ModuleNotFoundError:
    has_tk = False

# Stuff for terminal mode
enter_keys = [ord('\n'),ord('\r')]
panel_width = 80

source_url = 'https://github.com/USNavalResearchLaboratory/turboWAVE.git'
incompatibility_matrix = {
    'Windows' : ['GNU','Cray','CUDA','Apple CL','AMD ROCm','Homebrew','MacPorts'],
    'Linux' : ['Cray','Apple CL','Homebrew','MacPorts'],
    'MacOS' : ['Cray','Intel','CUDA','AMD ROCm','None'],
    'Cray' : ['GNU','LLVM','Apple CL','AMD ROCm','Homebrew','MacPorts'],
    'OpenMPI Cluster' : ['Cray','Apple CL','Homebrew','MacPorts'],
    'IntelMPI Cluster' : ['Cray','Apple CL','Homebrew','MacPorts']
}
help_str = '''This program configures, compiles, and installs turboWAVE core.
Press the buttons in the "Tasks" frame in order from top to bottom.
This will lead you through the necessary steps.'''
config_affirm='''The dropdown menus in the "Current Configuration" frame can be used to edit the configuration.
We have tried to choose the best options by detecting the environment.
Are you satisfied with the selections?'''

###########################################################
### BASE CLASSES FOR EITHER TEXT OR GRAPHICAL INTERFACE ###
###########################################################

def git_get_default_tag(package_path,taglist):
    '''Checks to see if the workspace is identical to any tagged commit.
    If it is return that tag, otherwise return "untagged".'''
    ans = 'untagged'
    save_dir = os.getcwd()
    os.chdir(package_path)
    for tag in taglist:
        compl = subprocess.run(["git","diff",tag],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
        if len(compl.stdout)<2:
            ans = tag
    os.chdir(save_dir)
    return ans

def git_is_clean(package_path):
    save_dir = os.getcwd()
    os.chdir(package_path)
    compl = subprocess.run(["git","status"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
    os.chdir(save_dir)
    return 'working tree clean' in compl.stdout.lower()

def git_is_detached(package_path):
    save_dir = os.getcwd()
    os.chdir(package_path)
    compl = subprocess.run(["git","status"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
    os.chdir(save_dir)
    return 'head detached' in compl.stdout.lower()

def git_err_handler(compl,cmd):
    if compl.returncode!=0:
        match = re.search(r'\b(err|ERR|Err).*\n',compl.stdout)
        if match:
            err_str = match.group(0)
        else:
            err_str = '(we cannot find the error string)'
        if cmd.affirm('Git returned an error:\n'+err_str+'\nWould you like to continue anyway?')=='y':
            return False
        else:
            return True
    else:
        return False

class dictionary_view:
    def __init__(self):
        self.highlighted = -1
        self.trouble = []
        self.title = 'Selection'
        self.data = {}
        self.choices = {}
        self.unpretty = {}
        self.widgets = {}
        self.selectable = []
    def highlight(self,i0):
        if i0==-1:
            self.highlighted = -1
            return
        i1 = 0
        for i,key in enumerate(self.data):
            if key in self.selectable:
                if i1==i0:
                    break
                i1 += 1
        self.highlighted = i
    def set_trouble(self,key):
        self.trouble += [key]
    def set_ok(self,key=None):
        if type(key)==type(None):
            self.trouble = []
        else:
            self.trouble = list(set(self.trouble)-{key})
    def get(self,key):
        return self.data[key]
    def verify(self):
        self.set_ok()
        return len(self.trouble)

class base_config(dictionary_view):
    def __init__(self):
        super().__init__()
        self.title = 'Current Configuration'
        self.unpretty['None'] = 'NONE'
        self.unpretty['Linux'] = 'LINUX'
        self.unpretty['Windows'] = 'WIN'
        self.unpretty['MacOS'] = 'OSX'
        self.unpretty['Cray'] = 'CRAY'
        self.unpretty['OpenMPI Cluster'] = 'OPENMPI'
        self.unpretty['IntelMPI Cluster'] = 'INTELMPI'
        self.unpretty['GNU'] = 'GNU'
        self.unpretty['LLVM'] = 'LLVM_CLANG'
        self.unpretty['Intel'] = 'INTEL'
        self.unpretty['Cray'] = 'CRAY'
        self.unpretty['OpenMP'] = 'OMP'
        self.unpretty['CUDA'] = 'CUDA'
        self.unpretty['Apple CL'] = 'APPLE_CL'
        self.unpretty['AMD ROCm'] = 'RADEON_PRO'
        self.unpretty['Homebrew'] = 'HOMEBREW'
        self.unpretty['MacPorts'] = 'MACPORTS'
        self.unpretty['Scalar'] = '64'
        self.unpretty['SSE'] = '128'
        self.unpretty['SSE2'] = '128'
        self.unpretty['AVX'] = '256'
        self.unpretty['AVX2'] = '256'
        self.unpretty['AVX512'] = '512'
        self.unpretty['Little Endian'] = 'LITTLE'
        self.unpretty['Big Endian'] = 'BIG'
        self.choices['Platform'] = ['Linux','Windows','MacOS','Cray','OpenMPI Cluster','IntelMPI Cluster']
        self.choices['Compiler'] = ['GNU','LLVM','Intel','Cray']
        self.choices['Accelerator'] = ['OpenMP','CUDA','Apple CL','AMD ROCm']
        self.choices['Packager'] = ['None','Homebrew','MacPorts']
        # the list of vectors is assumed to be in order of ascending preference
        self.choices['Vectors'] = ['Scalar','SSE','SSE2','AVX','AVX2','AVX512']
        self.choices['Byte Order'] = ['Little Endian','Big Endian']
        self.data['Tools Version'] = pkg_resources.get_distribution('twutils').version
        self.data['Core Version'] = ''
        self.data['Build Path'] = ''
        self.data['Platform'] = self.choices['Platform'][0]
        self.data['Compiler'] = self.choices['Compiler'][0]
        self.data['Accelerator'] = self.choices['Accelerator'][0]
        self.data['Vectors'] = self.choices['Vectors'][0]
        self.data['Packager'] = self.choices['Packager'][0]
        self.data['Byte Order'] = self.choices['Byte Order'][0]
        self.selectable = ['Platform','Compiler','Accelerator','Vectors','Packager','Byte Order']
        # Try to initialize the configuration based on environment
        if 'linux' in sys.platform:
            self.data['Platform'] = 'Linux'
            self.data['Compiler'] = 'LLVM'
            compl = subprocess.run(['lscpu'],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
            for v in reversed(self.choices['Vectors']):
                if re.search(r'\b'+v,compl.stdout,flags=re.IGNORECASE)!=None:
                    self.data['Vectors'] = v
                    break
        if 'darwin' in sys.platform:
            self.data['Platform'] = 'MacOS'
            self.data['Compiler'] = 'GNU'
            self.data['Packager'] = 'Homebrew'
            compl = subprocess.run(['sysctl','machdep.cpu'],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
            for v in reversed(self.choices['Vectors']):
                if re.search(r'\b'+v,compl.stdout,flags=re.IGNORECASE)!=None:
                    self.data['Vectors'] = v
                    break
        if 'win32' in sys.platform:
            self.data['Platform'] = 'Windows'
            self.data['Compiler'] = 'LLVM'
            self.data['Vectors'] = 'AVX2'
        if sys.byteorder=='big':
            self.data['Byte Order'] = 'Big Endian'
    def connect(self,tasks):
        self.tasks = tasks
    def backup_params(self):
        self.backup = copy.copy(self.data)
    def restore_params(self):
        self.data = copy.copy(self.backup)
    def get_make_val(self,key):
        return self.unpretty[self.data[key]]
    def get_params(self):
        ans = []
        for key in self.data:
            if key in self.choices:
                ans += [key]
        return ans
    def get_choices(self,key):
        return self.choices[key]
    def select(self,key,choice):
        self.set(key,choice)
        self.tasks.dirty_config()
    def verify(self):
        self.set_ok()
        if self.data['Compiler'] in incompatibility_matrix[self.data['Platform']]:
            self.set_trouble('Compiler')
        if self.data['Packager'] in incompatibility_matrix[self.data['Platform']]:
            self.set_trouble('Packager')
        if self.data['Accelerator'] in incompatibility_matrix[self.data['Platform']]:
            self.set_trouble('Accelerator')
        return len(self.trouble)

class base_tasks(dictionary_view):
    def __init__(self):
        super().__init__()
        self.title = 'Tasks'
        self.highlighted = 0
        self.data['Get Components'] = 'not started'
        self.data['Set Version'] = 'not started'
        self.data['Configure Makefile'] = 'not started'
        self.data['Compile Code'] = 'not started'
        self.data['Install Files'] = 'not started'
        self.data['Cleanup'] = 'not started'
        self.selectable = [key for key in self.data]
    def connect(self,cmd,conf):
        self.cmd = cmd
        self.conf = conf
    def dirty_config(self):
        if self.get('Configure Makefile')=='done':
            self.set('Configure Makefile','incomplete')
        if self.get('Compile Code')=='done':
            self.set('Compile Code','incomplete')
        if self.get('Install Files')=='done':
            self.set('Install Files','incomplete')
        if self.get('Cleanup')=='done':
            self.set('Cleanup','incomplete')
    def finished(self):
        for key in self.data:
            if self.data[key]!='done':
                return False
        return True
    def verify(self):
        self.set_ok()
        for key in self.data:
            if self.data[key]=='incomplete':
                self.set_trouble(key)
        return len(self.trouble)
    def GetComponents(self):
        self.set('Get Components','incomplete')
        test = []
        primitive_path = self.AskDirectory()
        if type(primitive_path)==tuple:
            return
        self.build_path = pathlib.Path(primitive_path) / 'turboWAVE' / 'core' / 'build'
        self.package_path = pathlib.Path(primitive_path) / 'turboWAVE'
        verified_path = glob.glob(str(self.build_path))
        if len(verified_path)==0:
            self.build_path = pathlib.Path(primitive_path) / 'core' / 'build'
            self.package_path = pathlib.Path(primitive_path)
            verified_path = glob.glob(str(self.build_path))
        if len(verified_path)==0:
            ans = self.cmd.affirm('Path is '+primitive_path+'\nTW source not found, would you like to retrieve it?')
            if ans=='y':
                dest = str(pathlib.Path(primitive_path) / 'turboWAVE')
                self.cmd.display('Cloning repository...')
                compl = subprocess.run(["git","clone",source_url,dest],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
                self.build_path = pathlib.Path(dest) / 'core' / 'build'
                self.package_path = pathlib.Path(dest)
                verified_path = glob.glob(str(self.build_path))
                if git_err_handler(compl,self.cmd):
                    verified_path = []
                self.cmd.display('')
        else:
            if git_is_clean(self.package_path):
                ans = self.cmd.affirm('TW source found, and working tree clean.\nWould you like to pull the latest?')
                if ans=='y':
                    save_dir = os.getcwd()
                    os.chdir(verified_path[0])
                    self.cmd.display('Pulling from repository...')
                    compl = subprocess.run(["git","pull","origin","master"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
                    if git_err_handler(compl,self.cmd):
                        verified_path = []
                    compl = subprocess.run(["git","pull","origin","--tag"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
                    if git_err_handler(compl,self.cmd):
                        verified_path = []
                    self.cmd.display('')
                    os.chdir(save_dir)
            else:
                self.cmd.info('Warning','TW source found, but working tree not clean.\nYou can continue if you know what you are doing.')
        if len(verified_path)>0:
            self.conf.set('Build Path',verified_path[0])
            self.set('Get Components','done')
        else:
            self.conf.set('Build Path','')
            self.cmd.err('Get components did not succeed.')

    def StartSetVersion(self):
        if self.get('Get Components')!='done':
            self.cmd.err('Please complete <Get Components> first.')
            return
        self.set('Set Version','incomplete')
        save_dir = os.getcwd()
        os.chdir(self.package_path)
        compl = subprocess.run(["git","tag","-l"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
        os.chdir(save_dir)
        if git_err_handler(compl,self.cmd)==False:
            rawlist = compl.stdout.splitlines()
            stable = []
            major = []
            prerelease = []
            for item in rawlist:
                if 'v' in item:
                    major += [item]
                elif 'a' in item or 'b' in item or 'rc' in item:
                    prerelease += [item]
                else:
                    stable += [item]
            taglist = sorted(stable[-5:])[::-1] + sorted(prerelease[-5:])[::-1]
            self.default_tag = 'workspace (' + git_get_default_tag(self.package_path,taglist) + ')'
            taglist = [self.default_tag] + taglist
            self.popup(taglist,self.FinishSetVersion)
    def FinishSetVersion(self,tag):
        if tag=='escape':
            return
        if tag==self.default_tag:
            self.conf.set('Core Version',self.default_tag)
            self.set('Set Version','done')
        else:
            save_dir = os.getcwd()
            os.chdir(self.package_path)
            compl = subprocess.run(["git","checkout",tag],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
            os.chdir(save_dir)
            if git_err_handler(compl,self.cmd)==False:
                self.conf.set('Core Version',tag)
                self.set('Set Version','done')
                if git_is_detached(self.package_path):
                    self.cmd.info('Notice','Local repo in detached state.  This can be fixed during cleanup.')
        toolsv = self.conf.data['Tools Version'].split('.')
        corev = tag.split('.')
        if toolsv[0]!=corev[0] or toolsv[1]!=corev[1]:
            self.cmd.info('Notice','Core and tool version mismatch.  For best results match core and tools to within a minor version number.')

    def Configure(self):
        if self.get('Set Version')!='done':
            self.cmd.err('Please complete <Set Version> first.')
            return
        self.set('Configure Makefile','incomplete')
        if self.cmd.affirm(config_affirm)=='n':
            return
        if self.conf.verify()>0:
            self.cmd.err('There are features selected that are incompatible with the platform.')
            return
        maker = 'GNU'
        if self.conf.get('Platform')=='Windows' and self.conf.get('Compiler')=='Intel':
            maker = 'MS'
        if maker=='MS':
            makefile_in = str(self.build_path / 'win.make')
            makefile_out = str(self.build_path / 'win.make.edited')
        if maker=='GNU':
            makefile_in = str(self.build_path / 'makefile')
            makefile_out = str(self.build_path / 'makefile.edited')
        with open(makefile_in,'r') as f:
            makefile = f.read()
        key_on = ['Platform','Compiler','Accelerator','Vectors','Packager','Byte Order']
        unpretty = {'Platform':'PLATFORM',
            'Compiler':'COMPILER_PREF',
            'Accelerator':'HARDWARE_ACCEL',
            'Vectors':'VBITS',
            'Packager':'PACKAGE_PREF',
            'Byte Order':'ENDIANNESS'}
        not_found = []
        for k in key_on:
            var = unpretty[k]
            val = self.conf.get_make_val(k)
            if re.search('\n[#\s]*' + var + '\s*=\s*',makefile)==None:
                not_found += [var]
            else:
                # Comment out all the values
                makefile = re.sub('\n[#\s]*'+var,'\n#'+var,makefile)
                # Don't care about preserving unused values, safest to just overwrite the first value found.
                makefile = re.sub('\n[#\s]*'+var+'\s*=\s*[\w\d]+','\n'+var+' = '+val,makefile,count=1)
        if self.conf.get('Packager')=='Homebrew' and self.conf.get('Compiler')=='GNU':
            gcc_glob_list = ['/usr/local/opt/**/g++-*','/opt/homebrew/**/g++-*']
            gcc_re = r'\/usr\/local\/opt\/gcc\/bin\/g\+\+-\d+'
            brew_gcc = []
            for gcc_glob in gcc_glob_list:
                brew_gcc += glob.glob(gcc_glob,recursive=True)
            if len(brew_gcc)==0:
                self.cmd.err('Did not find GCC of the form '+gcc_glob_list)
                return
            gcc_sel = brew_gcc[0]
            if len(brew_gcc)>1:
                brew_callback = lambda sel : (gcc_sel := sel,)[0]
                self.popup(brew_gcc,brew_callback)
            if re.search(gcc_re,makefile)==None:
                not_found += [gcc_sel]
            else:
                makefile = re.sub(gcc_re,gcc_sel,makefile)
        if len(not_found)==0:
            with open(makefile_out,'w') as f:
                f.write(makefile)
                self.set('Configure Makefile','done')
        else:
            missing = ''
            for s in not_found:
                missing += s + ','
            self.cmd.err('Did not find variables:'+missing[:-1]+'\nThis probably means core and tools are out of sync.\nYou might try a different version.')

    def Compile(self,enter_shell,leave_shell):
        if self.get('Configure Makefile')!='done':
            self.cmd.err('Please complete <Configure Makefile> first')
            return
        self.set('Compile Code','incomplete')
        if self.cmd.affirm('Compiler progress will appear in shell window.\nPlease wait for completion.\nReady to proceed?')=='y':
            self.cmd.display('Compiling...')
            save_dir = os.getcwd()
            os.chdir(self.build_path)
            maker = 'GNU'
            if self.conf.get('Platform')=='Windows' and self.conf.get('Compiler')=='Intel':
                maker = 'MS'
            enter_shell()
            try:
                if maker=='GNU':
                    compl = subprocess.run(['make','-f','makefile.edited','clean'],universal_newlines=True)
                    compl = subprocess.run(['make','-f','makefile.edited','tw3d_release'],universal_newlines=True)
                    if compl.returncode==0:
                        self.set('Compile Code','done')
                if maker=='MS':
                    compl = subprocess.run(['nmake','/F','win.make.edited','clean'],universal_newlines=True)
                    compl = subprocess.run(['nmake','/F','win.make.edited','tw3d'],universal_newlines=True)
                    if compl.returncode==0:
                        self.set('Compile Code','done')
                if self.get('Compile Code')!='done':
                    raise RuntimeError
            except FileNotFoundError:
                leave_shell()
                self.cmd.display('')
                if maker=='MS':
                    self.cmd.err('nmake is not supported in the synced core version.')#self.cmd.err('Could not find nmake. Try running twinstall from Intel shell.')
                else:
                    self.cmd.err('Could not find make, perhaps it is not installed?')
            except RuntimeError:
                leave_shell()
                self.cmd.display('')
                if maker=='MS':
                    self.cmd.err('nmake is not supported in the synced core version.')
                else:
                    self.cmd.err('There was an error while trying to compile.')
            except Exception as ex:
                leave_shell()
                self.cmd.display('')
                self.cmd.err('Unexpected error '+type(ex).__name__+', force exit.')
                exit(1)
            else:
                leave_shell()
                self.cmd.display('')
            os.chdir(save_dir)

    def Install(self):
        if self.get('Compile Code')!='done':
            self.cmd.err('Please complete <Compile Code> first.')
            return
        self.set('Install Files','incomplete')
        if self.conf.get('Platform')=='Windows':
            if self.conf.get('Compiler')=='Intel':
                src_exe = 'tw3d.exe'
            else:
                src_exe = 'tw3d'
            dst_exe = 'tw3d.exe'
            install_path = pathlib.Path(os.environ['CONDA_PREFIX']) / 'Scripts' / dst_exe
        else:
            src_exe = 'tw3d'
            dst_exe = 'tw3d'
            install_path = pathlib.Path(os.environ['CONDA_PREFIX']) / 'bin' / dst_exe
        repo_path = self.build_path / src_exe
        # tw3d executable
        if self.conf.get('Platform')=='Cray':
            dv_dir = self.AskDirectory('Select executable install location',str(pathlib.Path.home()))
            if type(dv_dir)!=tuple:
                shutil.copy(repo_path,dv_dir)
                self.set('Install Files','done')
        else:
            if self.cmd.affirm('We will install the main executable ('+dst_exe+') into '+str(install_path)+'. Proceed?')=='y':
                shutil.copy(repo_path,install_path)
                self.set('Install Files','done')
        # Jupyter DataViewer styles
        if self.cmd.affirm('Can we change Jupyter (increase notebook width)?')=='y':
            repo_path = self.package_path / 'tools' / 'config-files' / 'custom.css'
            dest_path = pathlib.Path.home() / '.jupyter' / 'custom'
            if len(glob.glob(str(dest_path)))==0:
                os.makedirs(dest_path,exist_ok=True)
            shutil.copy(repo_path,dest_path)
        # Jupyter DataViewer
        if self.cmd.affirm('Would you like make a copy of DataViewer?')=='y':
            repo_path = self.package_path / 'tools' / 'DataViewer.ipynb'
            dv_dir = self.AskDirectory('Select DataViewer install location',str(pathlib.Path.home()))
            if type(dv_dir)!=tuple:
                shutil.copy(repo_path,dv_dir)
        # MICRO highlighting
        if self.cmd.affirm('Would you like to install syntax highlights for MICRO?')=='y':
            repo_path = self.package_path / 'tools' / 'config-files' / 'turbowave.micro.yaml'
            dest_path = pathlib.Path.home() / '.config' / 'micro' / 'syntax'
            if len(glob.glob(str(dest_path)))==0:
                os.makedirs(dest_path,exist_ok=True)
            shutil.copy(repo_path,dest_path)
        # NANO highlighting
        if self.cmd.affirm('Would you like to install syntax highlights for NANO?')=='y':
            repo_path = self.package_path / 'tools' / 'config-files' / '.turbowave.nanorc'
            dest_path = pathlib.Path.home()
            shutil.copy(repo_path,dest_path)

            if self.conf.get('Platform')=='Windows':
                dest_path = pathlib.Path.home() / 'nano.rc'
            else:
                dest_path = pathlib.Path.home() / '.nanorc'
            try:
                with open(dest_path,'r') as f:
                    nanorc = f.read()
            except FileNotFoundError:
                nanorc = '# nano settings file, created by twinstall\n\n'
            inc_path = pathlib.Path.home() / '.turbowave.nanorc'
            search_str = r'include.+turbowave\.nanorc.*'
            inc_str = 'include ' + '"' + str(inc_path) + '"'
            if re.search(search_str,nanorc)==None:
                nanorc += '\n' + inc_str
            else:
                nanorc = re.sub(search_str,inc_str,nanorc)
            with open(dest_path,'w') as f:
                f.write(nanorc)
        # VIM highlighting
        if self.cmd.affirm('Would you like to install syntax highlights for VIM? (overwrites other filetype customizations)')=='y':
            repo_path = self.package_path / 'tools' / 'config-files' / 'filetype.vim'
            if self.conf.get('Platform')=='Windows':
                dest_path = pathlib.Path.home() / 'vimfiles'
            else:
                dest_path = pathlib.Path.home() / '.vim'
            if len(glob.glob(str(dest_path)))==0:
                os.makedirs(dest_path,exist_ok=True)
            shutil.copy(repo_path,dest_path)

            repo_path = self.package_path / 'tools' / 'config-files' / 'turbowave.vim'
            if self.conf.get('Platform')=='Windows':
                dest_path = pathlib.Path.home() / 'vimfiles' / 'syntax'
            else:
                dest_path = pathlib.Path.home() / '.vim' / 'syntax'
            if len(glob.glob(str(dest_path)))==0:
                os.makedirs(dest_path,exist_ok=True)
            shutil.copy(repo_path,dest_path)
        # Other highlighting
        self.cmd.info('Notice','Syntax highlights are available for Atom and VS Code, search for "turbowave" in the respective package managers.')

    def Clean(self):
        if self.get('Install Files')!='done':
            self.cmd.err('Please complete <Install Files> first.')
            return
        self.set('Cleanup','incomplete')
        if self.conf.get('Platform')=='Windows' and self.conf.get('Compiler')=='Intel':
            makefile_name = 'win.make.edited'
            obj_ext = '.obj'
            exe_name = 'tw3d.exe'
        else:
            makefile_name = 'makefile.edited'
            obj_ext = '.o'
            exe_name = 'tw3d'
        d = self.build_path / makefile_name
        if self.cmd.affirm('Delete '+str(d)+'?')=='y':
            os.remove(str(d))
        d = self.build_path / ('*' + obj_ext)
        if self.cmd.affirm('Delete '+str(d)+'?')=='y':
            obj_list = glob.glob(str(d))
            for f in obj_list:
                os.remove(f)
        d = self.build_path / exe_name
        if self.cmd.affirm('Delete '+str(d)+'?')=='y':
            os.remove(str(d))
        d = self.build_path / '*.d'
        if self.cmd.affirm('Delete '+str(d)+'?')=='y':
            os.remove(str(d))
        if git_is_detached(self.package_path):
            if self.cmd.affirm('Restore local repository to master?')=='y':
                save_dir = os.getcwd()
                os.chdir(self.package_path)
                compl = subprocess.run(["git","checkout","master"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
                os.chdir(save_dir)
                git_err_handler(compl,self.cmd)
        self.set('Cleanup','done')
        if '--terminal' in sys.argv or '-t' in sys.argv:
            self.cmd.display('All tasks are finished, you can press q to quit.')
        else:
            self.cmd.display('All tasks are finished, you can press <Quit>.')


######################
### GRAPHICAL MODE ###
######################

class gui_popup:
    def __init__(self,master_win,master_frame,items,completion_callback):
        self.window = tk.Toplevel()
        self.window.title('Selection')
        self.window.transient(master_frame)
        w = master_frame.winfo_width()
        h = master_frame.winfo_height()
        x = master_win.winfo_x() + master_frame.winfo_x()
        y = master_win.winfo_y() + master_frame.winfo_x()
        self.window.geometry('+'+str(int(w/4+x))+'+'+str(int(h/4+y)))
        self.window.grab_set()
        self.completion_callback = completion_callback
        mb = ttk.Menubutton(self.window,text='Select Version')
        mb.grid(column=0,row=0,sticky='ew',padx=5,pady=5)
        mb.menu = tk.Menu(mb,tearoff=0)
        mb['menu'] = mb.menu
        for choice in items:
            cmd = lambda choice=choice : self.select(choice)
            mb.menu.add_command(label=choice,command=cmd)
    def select(self,choice):
        self.completion_callback(choice)
        self.window.grab_release()
        self.window.destroy()

class gui_dictionary_view(dictionary_view):
    def __init__(self):
        super().__init__()
    def set(self,key,val):
        self.data[key] = val
        self.widgets[key][1].configure(text=val)
        self.verify()
        for key in self.data:
            if key in self.trouble:
                self.widgets[key][1].configure(foreground='red')
            else:
                self.widgets[key][1].configure(foreground='black')
    def popup(self,items,callback):
        return gui_popup(self.window.winfo_toplevel(),self.window,items,callback)

class gui_config(base_config,gui_dictionary_view):
    def __init__(self,master):
        super().__init__()
        self.window = ttk.LabelFrame(master,text=self.title)
        self.window.grid(row=0,column=0,padx=5,pady=5,sticky='ew')
        for i,key in enumerate(self.data):
            if key in self.choices:
                mb = ttk.Menubutton(self.window,text=key)
                mb.grid(column=0,row=i,sticky='ew',padx=5,pady=5)
                mb.menu = tk.Menu(mb,tearoff=0)
                mb['menu'] = mb.menu
                for choice in self.choices[key]:
                    cmd = lambda key=key,choice=choice : self.select(key,choice)
                    mb.menu.add_command(label=choice,command=cmd)
                var_label = ttk.Label(self.window,text=self.data[key],width=60,anchor='w')
                var_label.grid(column=1,row=i,padx=5,pady=5)
                self.widgets[key] = (mb,var_label)
            else:
                static_label = ttk.Label(self.window,text=key+':',anchor='e')
                static_label.grid(column=0,row=i,padx=5,pady=5,sticky='e')
                var_label = ttk.Label(self.window,text=self.data[key],width=60,anchor='w')
                var_label.grid(column=1,row=i,padx=5,pady=5)
                self.widgets[key] = (static_label,var_label)

class gui_tasks(base_tasks,gui_dictionary_view):
    def __init__(self,master):
        super().__init__()
        self.window = ttk.LabelFrame(master,text=self.title)
        self.window.grid(row=1,column=0,padx=5,pady=5,sticky='ew')
        for i,key in enumerate(self.data):
            button = ttk.Button(self.window,text=key)
            status = ttk.Label(self.window,text=self.data[key],width=60)
            button.grid(row=i,column=0,padx=5,pady=5,sticky='ew')
            status.grid(row=i,column=1,padx=5,pady=5)
            self.widgets[key] = (button,status)
        self.widgets['Get Components'][0].configure(command=self.GetComponents)
        self.widgets['Set Version'][0].configure(command=self.StartSetVersion)
        self.widgets['Configure Makefile'][0].configure(command=self.Configure)
        self.widgets['Compile Code'][0].configure(command=lambda : self.Compile(self.enter_shell,self.exit_shell))
        self.widgets['Install Files'][0].configure(command=self.Install)
        self.widgets['Cleanup'][0].configure(command=self.Clean)
    def enter_shell(self):
        print('Start Shell Output.')
    def exit_shell(self):
        print('Reached Completion. Stop Shell Output.')
    def AskDirectory(self,title='Select Working Directory',initialdir=str(pathlib.Path.home())):
        return tkinter.filedialog.askdirectory(title=title,initialdir=initialdir)

class gui_command:
    def __init__(self,master):
        self.window = ttk.Frame(master)
        self.window.grid(row=2,column=0,padx=5,pady=5)
        self.help = ttk.Button(self.window,text='Help',command=self.main_help)
        self.help.grid(row=0,column=0,padx=5,pady=5)
        self.quit = ttk.Button(self.window,text='Quit',command=master.quit)
        self.quit.grid(row=0,column=1,padx=5,pady=5)
        self.mess = ttk.Label(self.window)
        self.mess.grid(row=1,column=0,columnspan=2)
    def err(self,message):
        tk.messagebox.showerror(title='Notice',message=message)
    def affirm(self,message):
        affirmed = tk.messagebox.askyesno('Confirm',message)
        if affirmed:
            return 'y'
        else:
            return 'n'
    def display(self,message=''):
        self.mess.configure(text=message,font=('Times','14','bold'))
        self.mess.update()
    def info(self,title,message):
        tk.messagebox.showinfo(title=title,message=message)
    def main_help(self):
        tk.messagebox.showinfo(title='TurboWAVE Installer Help',message=help_str)

class Application(tk.Frame):
    def __init__(self,master=None):
        super().__init__(master)
        self.master = master
        master.title('TurboWAVE Core Installer')
        self.pack()
        self.cmd = gui_command(self)
        self.conf = gui_config(self)
        self.tasks = gui_tasks(self)
        self.tasks.connect(self.cmd,self.conf)
        self.conf.connect(self.tasks)
        master.update_idletasks()
        ws = master.winfo_screenwidth()
        hs = master.winfo_screenheight()
        w = master.winfo_width()
        h = master.winfo_height()
        master.geometry('+%d+%d' % (ws/2-w/2,hs/2-h/2))
        tst = self.conf.data['Tools Version']
        if 'a' in tst or 'b' in tst or 'rc' in tst:
            confirm = self.cmd.affirm('Tools version is unstable, proceed anyway')
            if confirm=='n':
                self.master.destroy()
    def quit(self):
        if not self.tasks.finished():
            confirm = self.cmd.affirm('There are incomplete tasks, do you want to quit?')
            if confirm=='n':
                return
        self.master.destroy()

#####################
### TERMINAL MODE ###
#####################

class term_titlebar:
    def __init__(self,title,y=0,x=0):
        self.title = title
        self.window = curses.newwin(1,panel_width,y,x)
    def display(self):
        self.window.clear()
        self.window.addstr(0,0,self.title,curses.A_BOLD)
        self.window.refresh()

class term_popup:
    def __init__(self,items,y=1,x=1):
        self.title = 'Select'
        self.highlighted = 0
        lens = [len(s) for s in items]
        self.dims = (2+len(items),2+max(lens+[len(self.title)]))
        self.window = curses.newwin(self.dims[0],self.dims[1],y,x)
        self.choices = items
    def getyx(self):
        return self.dims
    def display(self):
        self.window.clear()
        self.window.border()
        self.window.addstr(0,int(self.window.getmaxyx()[1]/2-len(self.title)/2),self.title,curses.A_BOLD)
        for i,item in enumerate(self.choices):
            if i==self.highlighted:
                self.window.addstr(i+1,1,item,curses.A_REVERSE)
            else:
                self.window.addstr(i+1,1,item)
        self.window.keypad(True)
        self.window.refresh()
    def keypress(self,c):
        N = len(self.choices)
        selection = -1
        if c==ord('q') or c==ord('Q'):
            return 'escape'
        if c==27:
            self.window.nodelay(True)
            if self.window.getch()==-1:
                return 'escape'
            self.window.nodelay(False)
        if self.highlighted==-1:
            self.highlighted = 0
        else:
            if c==curses.KEY_LEFT or c==curses.KEY_UP:
                self.highlighted -= 1
                if self.highlighted<0:
                    self.highlighted = N-1
            if c==curses.KEY_RIGHT or c==curses.KEY_DOWN:
                self.highlighted += 1
                if self.highlighted>=N:
                    self.highlighted = 0
            if c==ord('\n') or c==ord('\r'):
                selection = self.highlighted
        self.display()
        if selection==-1:
            return ''
        else:
            return self.choices[selection]
    def run(self):
        self.display()
        sel = ''
        while sel=='':
            c = self.window.getch()
            sel = self.keypress(c)
        return sel

class term_dictionary_view(dictionary_view):
    def __init__(self):
        super().__init__()
    def display(self):
        self.window.clear()
        self.window.border()
        self.window.addstr(0,int(self.window.getmaxyx()[1]/2-len(self.title)/2),self.title,curses.A_BOLD)
        for i,key in enumerate(self.data):
            line = str(i+1)+') '+key+' = '+self.data[key]
            if i==self.highlighted:
                if key in self.trouble:
                    self.window.addstr(i+1,1,line,curses.color_pair(1) | curses.A_REVERSE)
                else:
                    self.window.addstr(i+1,1,line,curses.A_REVERSE)
            else:
                if key in self.trouble:
                    self.window.addstr(i+1,1,line,curses.color_pair(1))
                else:
                    self.window.addstr(i+1,1,line)
        self.window.keypad(True)
        self.window.refresh()
    def set(self,key,val):
        self.data[key] = val
    def cyclic(self,i,N,delta,indices):
        i += delta
        if i<0:
            i = N-1
        if i>=N:
            i = 0
        while (i not in indices):
            i += delta
            if i<0:
                i = N-1
            if i>=N:
                i = 0
        return i
    def keypress(self,c):
        selection = -1
        max_index = len(self.data)
        indices = []
        for i,key in enumerate(self.data):
            if key in self.selectable:
                indices += [i]
        if self.highlighted==-1:
            self.highlighted = indices[0]
        else:
            if c==curses.KEY_LEFT or c==curses.KEY_UP:
                self.highlighted = self.cyclic(self.highlighted,max_index,-1,indices)
            if c==curses.KEY_RIGHT or c==curses.KEY_DOWN:
                self.highlighted = self.cyclic(self.highlighted,max_index,1,indices)
            if c==ord('\n') or c==ord('\r'):
                selection = self.highlighted
        self.display()
        if selection==-1:
            return ''
        else:
            return list(self.data.keys())[selection]
    def popup(self,items,callback):
        obj = term_popup(items)
        choice = obj.run()
        if choice!='escape':
            callback(choice)
        return choice

class term_config(base_config,term_dictionary_view):
    def __init__(self,y):
        super().__init__()
        self.window = curses.newwin(len(self.data)+2,panel_width,y,0)
    def rows(self):
        return len(self.data)+2
    def display(self):
        self.verify()
        super().display()

class term_tasks(base_tasks,term_dictionary_view):
    def __init__(self,y):
        super().__init__()
        self.window = curses.newwin(len(self.data)+2,panel_width,y,0)
        self.highlighted = 0
    def rows(self):
        return len(self.data)+2
    def display(self):
        self.set_ok()
        for key in self.data:
            if self.data[key]=='incomplete':
                self.set_trouble(key)
        super().display()
    def enter_shell(self):
        curses.endwin()
    def exit_shell(self):
        newscr = curses.initscr()
        newscr.refresh()
        curses.doupdate()
    def AskDirectory(self,title='Select Working Directory',initialdir=str(pathlib.Path.home())):
        self.cmd.display(title+'\nMust give complete path (no tab completion).\nEnvironment variables and ~ OK to use.')
        curses.echo()
        ans = self.cmd.getstr(3,0)
        curses.noecho()
        ans = ans.replace('~',str(pathlib.Path.home()))
        pattern = r'\$(\w+)'
        match = re.search(pattern,ans)
        while match!=None:
            ans = re.sub(pattern,os.environ[match[1]],ans,count=1)
            match = re.search(pattern,ans)
        if len(glob.glob(ans))==0:
            self.cmd.err('Could not find path.')
            ans = ()
        else:
            self.cmd.display('')
        return ans

class term_command:
    def __init__(self,y):
        self.window = curses.newwin(5,panel_width,y,0)
        self.default_mess = 'arrows=highlight, enter=select, q=exit, t=toggle'
        self.last_mess = 'arrows=highlight, enter=select, q=exit, t=toggle'
    def connect(self,conf,tasks,titlebar):
        self.conf = conf
        self.tasks = tasks
        self.titlebar = titlebar
    def display(self,message):
        self.window.clear()
        if message=='':
            self.window.addstr(0,0,self.default_mess)
            self.last_mess = self.default_mess
        elif message=='persist':
            self.window.addstr(0,0,self.last_mess)
        else:
            self.window.addstr(0,0,message)
            self.last_mess = message
        self.window.keypad(True)
        self.window.refresh()
    def err(self,message):
        self.display(message+'\n'+self.default_mess)
    def info(self,title,message):
        self.display(message+' (press any key)')
        self.window.getch()
        self.display('')
    def affirm(self,message):
        self.display(message+' (y/n)?')
        confirm = ord('x')
        while confirm not in [ord('y'),ord('n')]:
            confirm = self.window.getch()
        self.display('')
        return chr(confirm)
    def getstr(self,y,x):
        curses.curs_set(1)
        s = str(self.window.getstr(y,x),encoding='utf-8')
        curses.curs_set(0)
        return s
    def run(self):
        key_focus = self.tasks
        tst = self.conf.data['Tools Version']
        if 'a' in tst or 'b' in tst or 'rc' in tst:
            confirm = self.affirm('Tools version is unstable, proceed anyway')
            if confirm=='n':
                return
        while True:
            self.titlebar.display()
            self.conf.display()
            self.tasks.display()
            self.display('persist')
            c = self.window.getch()
            if c==ord('q'):
                if self.tasks.finished():
                    break
                confirm = self.affirm('There are incomplete tasks, do you want to quit')
                if confirm=='y':
                    break
            if c==ord('t'):
                if key_focus==self.tasks:
                    key_focus = self.conf
                    self.conf.highlight(0)
                    self.tasks.highlight(-1)
                else:
                    key_focus = self.tasks
                    self.tasks.highlight(0)
                    self.conf.highlight(-1)
            sel = key_focus.keypress(c)
            if sel in self.conf.get_params():
                obj = term_popup(self.conf.get_choices(sel))
                res = obj.run()
                if res!='escape':
                    self.tasks.dirty_config()
                    self.conf.set(sel,res)
            if sel=='Get Components':
                self.tasks.GetComponents()
            if sel=='Set Version':
                self.tasks.StartSetVersion()
            if sel=='Configure Makefile':
                self.tasks.Configure()
            if sel=='Compile Code':
                self.tasks.Compile(self.tasks.enter_shell,self.tasks.exit_shell)
            if sel=='Install Files':
                self.tasks.Install()
            if sel=='Cleanup':
                self.tasks.Clean()

def terminal(stdscr):
    curses.init_pair(1,curses.COLOR_RED,curses.COLOR_BLACK)
    stdscr.clear()
    curses.curs_set(0) # make cursor invisible

    titlebar = term_titlebar('TurboWAVE Core Installer')
    conf = term_config(1)
    tasks = term_tasks(1+conf.rows())
    cmd = term_command(1+conf.rows()+tasks.rows())
    conf.connect(tasks)
    tasks.connect(cmd,conf)
    cmd.connect(conf,tasks,titlebar)
    cmd.run()

def main():
    if '--help' in sys.argv or '-h' in sys.argv:
        print('TurboWAVE Interactive Core Installer.')
        print('Usage: twinstall [--help] [--terminal]')
        print('Arguments: --help,-h :  displays this message')
        print('           --version,-v : display version number')
        print('           --terminal,-t : use textual interface')
        exit(0)

    if '--version' in sys.argv or '-v' in sys.argv:
        print('twinstall is provided by twutils, version '+pkg_resources.get_distribution('twutils').version)
        exit(0)

    if '--terminal' in sys.argv or '-t' in sys.argv:
        os.environ.setdefault('ESCDELAY','25')
        curses.wrapper(terminal)

    else:
        root = tk.Tk()
        ttk.Style().theme_use('default')
        app = Application(master=root)
        app.mainloop()



'''
Module :samp:`twutils.NLO`
====================================

This module contains parameters for dielectric materials that can be used in turboWAVE, as well as functions to evaluate their numerical dispersion properties.  This module uses mks units.

Classes and functions
----------------------
'''

import numpy as np
import scipy as sp
from scipy import constants as C

class vacuum:
	"""Contains parameters describing vacuum"""
	num = 0
	Omega = []
	wp = []
	gamma = []
	f = []
	a = []
	b = []
	d = []

class fused_silica:
	"""Contains parameters describing 3 oscillator model for fused silica"""
	num = 3
	Omega = [[2.346e14]*3 , [1.627e16]*3 , [2.776e16]*3]
	wp = [1.794e14 , 1.055e16 , 2.296e16]
	gamma = [[0.0]*3]*3
	f = [[1.0]*3]*3
	a = [[[0.0]*6]*3]*3
	b = [0.0]*3
	d = [0.0]*3

class mag_fluoride:
	"""Contains parameters describing 3 oscillator model for magnesium fluoride"""
	num = 3
	Omega = [4.342e16 , 1.991e16 , 7.917e13]
	wp = [3.032e16 , 1.257e16 , 1.204e14]
	gamma = [[0.0]*3]*3
	f = [[1.0]*3]*3
	a = [[[0.0]*6]*3]*3
	b = [0.0]*3
	d = [0.0]*3

class silicon_carbide:
	"""Contains parameters describing 2 oscillator model for silicon carbide"""
	num = 2
	Omega = [[1.828e14]*3 , [1.885e15]*3]
	wp = [2.733e14 , 4.524e15]
	gamma = [[8.954e11*0.5]*3 , [0.0]*3]
	f = [[1.0]*3]*2
	a = [[[0.0]*6]*3]*2
	b = [0.0]*2
	d = [0.0]*2

class BBO:
	"""Contains parameters describing 2 oscillator model for BBO crystals"""
	num = 2
	Omega = [[1.714e16,1.714e16,1.880e16],[1.590e15,1.590e15,1.356e15]]
	wp = [1e16,1e14]
	gamma = [[0.0]*3]*2
	f = [[5.075,5.075,4.833],[0.6823,0.6823,1.653]]
	a = [
		[[-0.00e+00, -0.00e+00, -0.00e+00, -0.00e+00, -4.89e+40, -1.11e+42],
		[-1.11e+42,  1.11e+42, -0.00e+00, -4.89e+40, -0.00e+00, -0.00e+00],
		[-2.59e+40, -2.59e+40, -3.77e+40, -0.00e+00, -0.00e+00, -0.00e+00]],
		[[0.0]*6]*3
		]
	b = [1.6e54,0.0]
	d = [-7.2e74,0.0]


def analyticWavenumber(w,oscParams,n=1):
	"""Compute k(w) from analytical dispersion relation

	:param float w: angular frequency [mks]
	:param class oscParams: instance of an oscillator class defined in this module
	:param int n: polarization axis numbered from 1-3"""
	ans = w**2 + 0j
	for i in range(oscParams.num):
		ans -= (oscParams.f[i][n]*oscParams.wp[i]**2 * w**2) / (w**2 + 2j*oscParams.gamma[i][n]*w - oscParams.Omega[i][n]**2)
	return np.sqrt(ans)/C.c

def numericalWavenumber(w,oscParams,dt,dz,n=1):
	"""Compute k(w) from numerical dispersion relation

	:param float w: angular frequency [mks]
	:param class oscParams: instance of an oscillator class defined in this module
	:param float dt: timestep [mks]
	:param float dz: cell size [mks]
	:param int n: polarization axis numbered from 1-3"""
	Pol = 0.0
	for i in range(oscParams.num):
		f = oscParams.f[i][n]
		wp = oscParams.wp[i]
		W = oscParams.Omega[i][n]
		g = oscParams.gamma[i][n]
		denom = (1.0+g*dt)*np.sin(0.5*w*dt)**2 - 0.25*(W*dt)**2 + 0.5j*g*dt*np.sin(w*dt)
		Pol += (0.5*f*(wp*dt)**2 * np.sin(0.5*w*dt)**2)/denom
	Vac_deduced = np.cos(w*dt) + Pol
	sin_factor = np.sqrt(0.5*(1.0-Vac_deduced)*dz**2/(C.c*dt)**2)
	return np.arcsin(sin_factor)*2.0/dz

def analyticDF(wa,kz,oscParams,n=1):
	"""Returns analytical dispersion function (intended as argument to ``fsolve``)

	:param array-like wa: two-element array with real & imaginary part of angular frequency
	:param float kz: real wavenumber
	:param class oscParams: instance of an oscillator class defined in this module
	:param int n: polarization axis numbered from 1-3"""
	w = wa[0] + 1j*wa[1]
	ans = w**2 - (C.c*kz)**2
	for i in range(oscParams.num):
		ans -= (oscParams.f[i][n]*oscParams.wp[i]**2 * w**2) / (w**2 + 2j*oscParams.gamma[i][n]*w - oscParams.Omega[i][n]**2)
	return [ans.real,ans.imag]

def numericalDF(wa,kz,oscParams,dt,dz,n=1):
	"""Returns numerical dispersion function (intended as argument to ``fsolve``)

	:param array-like wa: two-element array with real & imaginary part of angular frequency
	:param float kz: real wavenumber
	:param class oscParams: instance of an oscillator class defined in this module
	:param float dt: timestep [mks]
	:param float dz: cell size [mks]
	:param int n: polarization axis numbered from 1-3"""
	w = wa[0] + 1j*wa[1]
	Vac = 1.0 - 2.0 * (C.c*dt)**2 * (np.sin(0.5*kz*dz)**2 / dz**2)
	Pol = 0.0
	for i in range(oscParams.num):
		f = oscParams.f[i][n]
		wp = oscParams.wp[i]
		W = oscParams.Omega[i][n]
		g = oscParams.gamma[i][n]
		denom = (1.0+g*dt)*np.sin(0.5*w*dt)**2 - 0.25*(W*dt)**2 + 0.5j*g*dt*np.sin(w*dt)
		Pol += (0.5*f*(wp*dt)**2 * np.sin(0.5*w*dt)**2)/denom
	ans = np.cos(w*dt) - Vac + Pol
	return [ans.real,ans.imag]



import sys
import glob
import os
import pkg_resources
import PIL.Image
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import twutils.plot as twplot
from scipy import constants as C
eta0 = np.sqrt(C.mu_0/C.epsilon_0)

def print_usage():
    print('Usage: wigner slicing=slices/... real_file,... imag_file,... [panels=a,b,...] [layout=2x2]')
    print('   [dr=0.0,0.0,...] [color=viridis,jet,...] [roi=h0,h1,v0,v1/...]')
    print('   [range=low,high/...] [type=<type1>,...]')
    print('------------------Examples----------------')
    print('Envelope: python wigner.py xyzt=0,0,0 e_real.npy e_imag.npy')
    print('Carrier resolved: python wigner.py xyzt=0,0,0 Ex.npy 1.0')
    print('-------------------General Notes--------------------')
    print('Extra spaces (e.g. around commas, semicolons, or equals) are not allowed.')
    print('Displays two panels, the field in real space, and the field in Wigner phase space.')
    print('Required arguments are positional, optional arguments are key=value pairs.')
    print('Double values refer to the two plots.')
    print('------------------Arguments-----------------------')
    print('slicing: 4-character string, such as xyzt, where the first axis is the one to transform.')
    print('slices: is a comma delimited list of 2 (pulse) or 3 (spectrum,wigner) slice indices.')
    print('real_file: name of the file with the real part of the field')
    print('imag_file: EITHER name of the file with the imaginary part of the field')
    print('   or the carrier frequency if the data is real.')
    print('layout: grid to layout the plots on')
    print('dr: 0.0 signals full range on linear scale, any other float is the number of decades spanned.')
    print('color: viridis,magma,plasma,inferno,Spectral,bwr,seismic,prism,ocean,rainbow,jet,nipy_spectral')
    print('   Color maps may be inverted by adding "_r" to the name')
    print('   Note any Matplotlib color maps can be used.')
    print('roi: select a subdomain to plot, otherwise the full domain is plotted.')
    print('range: range of the color bar data.')
    print('type: can be pulse,spectrum,wigner')
    print('----------------------Animations----------------------')
    print('Put a python range as one of the slices to generate animated GIF.')
    print('For example, zxyt=0,0,2:5 would animate time slices 2,3,4.')

def print_version():
    print('wigner is provided by twutils, version '+pkg_resources.get_distribution('twutils').version)

def WignerTransform(A,ds,eta0):
    # If A is in V/m, ds in secs, and eta0 in ohms, returns J/m^2.
    # Then, integration over consistent phase space units (dimensionless product) gives the actual fluence.
    # The frequency variable is assumed to be an angular frequency (e.g. rad/s)
    N = A.shape[0]
    M = np.int(N/2) + 1
    corr = np.zeros((N,M)).astype(np.complex)
    Ai = np.zeros(N*2-1).astype(np.complex)
    Ai[::2] = A
    Ai[1::2] = 0.5*(np.roll(Ai,1)+np.roll(Ai,-1))[1::2]
    for j in range(M):
        corr[:,j] = (np.conj(np.roll(Ai,j))*np.roll(Ai,-j))[::2]
    wig = np.fft.hfft(corr,axis=1)*ds/(2*np.pi)
    wig = np.fft.fftshift(wig,axes=1)
    # Fix the units
    dw = (2*np.pi/ds) / N
    fluence = 0.5*np.sum(np.abs(A)**2)*ds/eta0
    wigfluence = np.sum(wig)*ds*dw
    return wig*fluence/wigfluence

def cleanup(wildcarded_path):
    cleanstr = glob.glob(wildcarded_path)
    for f in cleanstr:
        os.remove(f)

def ParseSlices(dims,ax_list,slice_str_list):
    '''Function to generate a list of slice tuples for the movie.
    dims = dimensions of all 4 axes
    ax_list = slicing_spec as list of integer axis identifiers.
    slice_str_list = list of slice strings, can be indices or ranges.
    Returns slice_tuples,movie.'''
    slice_tuples = []
    range_tuples = []
    movie = False
    # Construct list of range tuples
    sax = ax_list[4-len(slice_str_list):]
    for saxi,slice_str in enumerate(slice_str_list):
        rng = slice_str.split(':')
        tup = ()
        for i,el in enumerate(rng):
            if el=='' and i==0:
                el = '0'
            if el=='' and i==1:
                el = str(dims[sax[saxi]])
            if el=='' and i==2:
                el = '1'
            tup += (int(el),)
        range_tuples.append(tup)
    # Determine the range of the movie frames
    frame_rng = range(1)
    for rng in range_tuples:
        movie = movie or len(rng)>1
        if len(rng)==2:
            frame_rng = range(rng[0],rng[1])
        if len(rng)==3:
            frame_rng = range(rng[0],rng[1],rng[2])
    # Construct list of slice tuples
    for r in frame_rng:
        tup = ()
        for rng in range_tuples:
            if len(rng)>1:
                tup += (r,)
            else:
                tup += rng
        slice_tuples.append(tup)
    return slice_tuples,movie

def get_plotters(real_data_file,imag_data_file):
    plotter_r = twplot.plotter(real_data_file,buffered=False,units='mks')
    try:
        carrier = float(imag_data_file)
        plotter_i = 0.0
    except ValueError:
        carrier = 0.0
        plotter_i = twplot.plotter(imag_data_file,buffered=False,units='mks')
    return plotter_r,plotter_i,carrier

def form_envelope(real_field,carrier,dz,ax):
    k = 2*np.pi*np.fft.fftfreq(real_field.shape[ax],dz)
    dk = k[1]-k[0]
    kc = -k[int(real_field.shape[ax]/2)]
    carrier_idx = int(real_field.shape[ax] * carrier / (2*C.c*kc))
    env = np.fft.fft(real_field,axis=ax)
    if ax==0:
        env[int(env.shape[0]/2):,...] = 0.0
    else:
        env[...,int(env.shape[1]/2):] = 0.0
    env = np.roll(env,-carrier_idx,axis=ax)
    return 2*np.fft.ifft(env,axis=ax)

def get_spectrum(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range):
    if carrier!=0.0:
        abcissa,real1d,dict1d = plotter_r.lineout(slicing_spec,slice_now,dyn_range)
        envelope1d = form_envelope(real1d,carrier,abcissa[1]-abcissa[0],0)
    else:
        abcissa,real1d,dict1d = plotter_r.lineout(slicing_spec,slice_now,dyn_range)
        abcissa,imag1d,dict1d = plotter_i.lineout(slicing_spec,slice_now,dyn_range)
        envelope1d = (real1d + 1j*imag1d)
    spectrum = np.fft.fftshift(np.abs(np.fft.fft(envelope1d))**2)
    dwpts = np.fft.fftshift(2*np.pi*C.c*np.fft.fftfreq(envelope1d.shape[0],abcissa[1]-abcissa[0]))
    return spectrum,dwpts

def get_pulse(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range):
    if carrier!=0.0:
        real2d,dict2d = plotter_r.falsecolor2d(slicing_spec,slice_now[1:],dyn_range)
        abcissa,real1d,dict1d = plotter_r.lineout(slicing_spec,slice_now,dyn_range)
        envelope2d = form_envelope(real2d,carrier,abcissa[1]-abcissa[0],1)
    else:
        real2d,dict2d = plotter_r.falsecolor2d(slicing_spec,slice_now[1:],dyn_range)
        imag2d,dict2d = plotter_i.falsecolor2d(slicing_spec,slice_now[1:],dyn_range)
        envelope2d = (real2d + 1j*imag2d)
    # Put result in TV/cm
    return np.abs(envelope2d)*1e-12*1e-2,dict2d['extent']

def get_wigner(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range):
    if carrier!=0.0:
        abcissa,real1d,dict1d = plotter_r.lineout(slicing_spec,slice_now,dyn_range)
        envelope1d = form_envelope(real1d,carrier,abcissa[1]-abcissa[0],0)
    else:
        abcissa,real1d,dict1d = plotter_r.lineout(slicing_spec,slice_now,dyn_range)
        abcissa,imag1d,dict1d = plotter_i.lineout(slicing_spec,slice_now,dyn_range)
        envelope1d = (real1d + 1j*imag1d)
    z_extent = [abcissa[0],abcissa[-1]]
    dz = (z_extent[1]-z_extent[0]) / envelope1d.shape[0]
    k_extent = [-np.pi*C.c/dz,np.pi*C.c/dz]
    wig_ext = z_extent + k_extent
    # Put result in MJ/cm2
    return WignerTransform(envelope1d,dz/C.c,eta0)*1e-6*1e-4,wig_ext

def get_plot_data(plotter_r,plotter_i,carrier,plot_type,slicing_spec,slice_now,dyn_range):
    if plot_type=='pulse':
        return get_pulse(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range)
    if plot_type=='spectrum':
        return get_spectrum(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range)
    if plot_type=='wigner':
        return get_wigner(plotter_r,plotter_i,carrier,slicing_spec,slice_now,dyn_range)

def main():

    if len(sys.argv)<2 or '--help' in sys.argv or '-h' in sys.argv:
        print_usage()
        exit(0)

    if '--version' in sys.argv or '-v' in sys.argv:
        print_version()
        exit(0)

    # Matplotlib setup and default args

    mpl.rcParams.update({'text.usetex' : False , 'font.size' : 10})
    proportional = False
    if proportional:
        my_aspect = 'equal'
    else:
        my_aspect = 'auto'

    # Required arguments

    slicing_spec = []
    primitive_slices = []
    real_data_file = []
    imag_data_file = []
    for s in sys.argv[1].split('/'):
        slicing_spec += [s.split('=')[0]]
        primitive_slices += [(s.split('=')[1]).split(',')]
    for f in sys.argv[2].split(','):
        real_data_file += [f]
    for f in sys.argv[3].split(','):
        imag_data_file += [f]
    N = len(real_data_file)
    if len(imag_data_file)!=N:
        raise ValueError('Number of files does not match.')

    # Optional arguments

    color = []
    dyn_range = []
    roi = []
    val_range = []
    ask = 'yes'
    panels = []
    type = []
    cols = 1
    rows = N
    for keyval in sys.argv[4:]:
        key = keyval.split('=')[0]
        val = keyval.split('=')[1]
        if key=='panels':
            panels = val.split(',')
        if key=='layout':
            layout = val.split('x')
            rows = int(layout[0])
            cols = int(layout[1])
        if key=='dr':
            for dr in val.split(','):
                dyn_range += [float(dr)]
        if key=='color':
            color = val.split(',')
        if key=='roi':
            roi = val.split('/')
        if key=='range':
            val_range = val.split('/')
        if key=='type':
            type = val.split(',')

    # Fill in defaults
    for i in range(N-len(slicing_spec)):
        slicing_spec.append(slicing_spec[-1])
        primitive_slices.append(primitive_slices[-1])
    for i in range(N-len(panels)):
        panels.append('')
    for i in range(N-len(dyn_range)):
        dyn_range.append(0.0)
    for i in range(N-len(color)):
        color.append('viridis')
    for i in range(N-len(roi)):
        roi.append('')
    for i in range(N-len(val_range)):
        val_range.append('')
    for i in range(N-len(type)):
        type.append('wigner')

    # Check existing image files and clean

    img_files = glob.glob('frame*.png')
    if len(img_files)>0 and ask=='yes':
        ans = ''
        while ans!='y' and ans!='n':
            ans = input('Found some frame*.png files, OK to clean (y/n) ?')
        if ans=='n':
            print('STOPPED. Please run script in a directory where there are no important files of the form frame*.png.')
            exit(1)

    for img_file in img_files:
        os.remove(img_file)


    # Determine the global color scale bounds for all plots
    # If a movie we have to do all the transforms first
    # We don't save the data, just do redundant transforms later

    num_frames = []
    global_min = [1e50]*N
    global_max = [-1e50]*N
    print('First pass...')
    print('-------------')
    for i in range(N):
        plotter_r,plotter_i,carrier = get_plotters(real_data_file[i],imag_data_file[i])
        axes = twplot.get_axis_info(slicing_spec[i])
        dims = plotter_r.dims4()
        slice_tuples,movie = ParseSlices(dims,axes,primitive_slices[i])
        num_frames += [len(slice_tuples)]

        if val_range[i]=='':
            for slice_now in slice_tuples:
                dat,ext = get_plot_data(plotter_r,plotter_i,carrier,type[i],slicing_spec[i],slice_now,dyn_range[i])
                local_min = np.min(dat)
                local_max = np.max(dat)
                if local_min<global_min[i]:
                    global_min[i] = local_min
                if local_max>global_max[i]:
                    global_max[i] = local_max
        else:
            global_min[i] = float(val_range[i].split(',')[0])
            global_max[i] = float(val_range[i].split(',')[1])

    # Make a movie or display single frame

    print('Generate images...')
    print('------------------')

    for file_idx in range(num_frames[0]):

        sz = (cols*5,rows*4)
        plt.figure(file_idx,figsize=sz,dpi=100)

        for i in range(N):
            plt.subplot(rows,cols,i+1)
            plotter_r,plotter_i,carrier = get_plotters(real_data_file[i],imag_data_file[i])
            axes = twplot.get_axis_info(slicing_spec[i])
            dims = plotter_r.dims4()
            slice_tuples,movie = ParseSlices(dims,axes,primitive_slices[i])
            slice_now = slice_tuples[file_idx]
            dat,ext = get_plot_data(plotter_r,plotter_i,carrier,type[i],slicing_spec[i],slice_now,dyn_range[i])

            if type[i]=='pulse':
                ext_scaled = list(1e15*np.array(ext[:2])/C.c) + list(1e6*np.array(ext[2:]))
                plt.imshow(dat,
                    origin='lower',
                    aspect=my_aspect,
                    extent=ext_scaled,
                    vmin=global_min[i],
                    vmax=global_max[i],
                    cmap=color[i])
                b = plt.colorbar()
                b.set_label(r'${\cal E}$ (TV/cm)',size=12)
                plt.xlabel(r'$z/c - t$ (fs)',size=12)
                plt.ylabel(r'$\rho$ ($\mu$m)',size=12)

            if type[i]=='wigner':
                ext_scaled = list(1e15*np.array(ext[:2])/C.c) + list(1e-15*np.array(ext[2:]))
                plt.imshow(dat.swapaxes(0,1),
                    origin='lower',
                    aspect=my_aspect,
                    extent=ext_scaled,
                    vmin=global_min[i],
                    vmax=global_max[i],
                    cmap=color[i])
                b = plt.colorbar()
                b.set_label(r'${\cal N}$ (MJ/cm$^2$)',size=12)
                plt.xlabel(r'$z/c - t$ (fs)',size=12)
                plt.ylabel(r'$\delta\omega$ (fs$^{-1}$)',size=12)

            if type[i]=='spectrum':
                ext *= 1e-15
                ext_scaled = [ext[0],ext[-1],global_min[i],global_max[i]]
                plt.plot(ext,dat)
                plt.xlabel(r'$\delta\omega$ (fs$^{-1}$)',size=12)
                plt.ylabel(r'$|E(\omega)|^2$',size=12)

            if roi[i]=='':
                roi_i = ext_scaled
            else:
                roi_i = []
                for s in roi[i].split(','):
                    roi_i.append(float(s))
                plt.xlim(roi_i[0],roi_i[1])
                plt.ylim(roi_i[2],roi_i[3])
            if not panels[i]=='':
                plt.text(roi_i[0],roi_i[3]+0.03*(roi_i[3]-roi_i[2]),'('+panels[i]+')')

        plt.tight_layout()

        if movie:
            img_file = 'frame{:03d}.png'.format(file_idx)
            print('saving',img_file,'...')
            plt.savefig(img_file)
            plt.close()

    if movie:
        print('Consolidating into movie file...')
        images = []
        frameRateHz = 5
        img_files = sorted(glob.glob('frame*.png'))
        for f in img_files:
            images.append(PIL.Image.open(f))
        images[0].save('mov.gif',save_all=True,append_images=images[1:],duration=int(1000/frameRateHz),loop=0)
        cleanup('frame*.png')
        print('Done.')
    else:
        plt.show()



"""Module to setup plots based on npy and accompanying metadata.
Primary interaction is through the plotter class."""

import sys
import os
import json
import warnings
import importlib
import numpy as np
from scipy import constants as C

def axis_num(axis_letter):
    if axis_letter=='t':
        return 0
    if axis_letter=='x':
        return 1
    if axis_letter=='y':
        return 2
    if axis_letter=='z':
        return 3

def axis_lett(axis_num):
    if axis_num==0:
        return 't'
    if axis_num==1:
        return 'x'
    if axis_num==2:
        return 'y'
    if axis_num==3:
        return 'z'

def expand_slice_spec(primitive_spec):
    """Given a 2-character plane of interest add the letters for the remaining 2 axes.
    The order is such that time will never be last."""
    h = axis_num(primitive_spec[0])
    v = axis_num(primitive_spec[1])
    s1 = 3
    while s1==h or s1==v:
        s1 = s1 - 1
    s2 = 3
    while s2==h or s2==v or s2==s1:
        s2 = s2 - 1
    return primitive_spec + axis_lett(s2) + axis_lett(s1)

def get_axis_info(slicing_spec):
    """Get an ordered array of axis indices from a slicing string.
    Argument is a 4 character string giving the axis order (field component is always last).
    Return a 5 element axis array of integers, with encoding t=0,x=1,y=2,z=3,component=4.
    Example: 'zxyt' would give [3,1,2,0,4]."""
    axes = np.array([0,0,0,0,0]).astype(np.int)
    for i in range(4):
        axes[i] = axis_num(slicing_spec[i])
    axes[4] = 4
    return axes

def get_mesh_pts(grid_file_path,dims):
    """Try to find a grid warp file matching the npy file.
    If the file is found return [tmesh,xmesh,ymesh,zmesh].
    Each element is a list of the mesh points along the given axis.
    If the file is not found return unit length uniform mesh."""
    try:
        ans = [[],[],[],[]]
        with open(grid_file_path) as f:
            for line in f:
                l = line.split(' ')
                if l[0]=='t':
                    ans[0] += [np.float(l[-1])]
                if l[0]=='axis1' and len(ans[1])==0:
                    ans[1] = [np.float(x) for x in l[2:]]
                if l[0]=='axis2' and len(ans[2])==0:
                    ans[2] = [np.float(x) for x in l[2:]]
                if l[0]=='axis3' and len(ans[3])==0:
                    ans[3] = [np.float(x) for x in l[2:]]
        if len(ans[0])==dims[0] and len(ans[1])==dims[1] and len(ans[2])==dims[2] and len(ans[3])==dims[3]:
            return [np.array(ans[0]),np.array(ans[1]),np.array(ans[2]),np.array(ans[3])]
        else:
            warnings.warn('Grid file found but wrong dimensions ('+grid_file_path+').')
            return [np.linspace(0.0,1.0,dims[i]) for i in range(4)]
    except FileNotFoundError:
        warnings.warn('No grid file found ('+grid_file_path+').')
    except PermissionError:
        warnings.warn('Cannot open grid file ('+grid_file_path+') due to permissions.')
    return [np.linspace(0.0,1.0,dims[i]) for i in range(4)]

class plotter:
    """Class that retains information about turboWAVE data,
    and generates objects suitable for direct use in matplotlib."""
    def __init__(self,file_to_plot,units='plasma',buffered=True,small_pos=1e-25):
        self.basepath = os.path.split(file_to_plot)[0]
        self.name = os.path.split(file_to_plot)[1] # filename only
        self.path = file_to_plot # includes path
        self.units = units
        self.buffered = buffered
        self.meta = {}
        try:
            with open(os.path.join(self.basepath,'tw_metadata.json'),'r') as f:
                meta_str = f.read()
            try:
                self.meta = json.loads(meta_str)
            except ValueError:
                warnings.warn('Could not decode tw_metadata.json.')
        except (FileNotFoundError,PermissionError):
            warnings.warn('Could not read file '+os.path.join(self.basepath,'tw_metadata.json'))
        try:
            self.axes = self.meta[self.name]['axes']
            for i in range(5):
                self.axes[i] = self.axes[str(i)]
        except LookupError:
            warnings.warn('Could not retrieve axes from the metadata for '+self.name+', using arbitrary scales.')
            self.axes = { 0 : {} , 1 : {} , 2 : {} , 3 : {} , 4 : {} }
            self.axes[0]['label'] = 't'
            self.axes[1]['label'] = 'x'
            self.axes[2]['label'] = 'y'
            self.axes[3]['label'] = 'z'
            self.axes[4]['label'] = 'data'
            for i in range(5):
                self.axes[i]['mks label'] = 'None'
                self.axes[i]['mks multiplier'] = 1.0
                self.axes[i]['cgs label'] = 'None'
                self.axes[i]['cgs multiplier'] = 1.0
                self.axes[i]['plasma label'] = 'None'
                self.axes[i]['plasma multiplier'] = 1.0
                self.axes[i]['atomic label'] = 'None'
                self.axes[i]['atomic multiplier'] = 1.0
                self.axes[i]['natural label'] = 'None'
                self.axes[i]['natural multiplier'] = 1.0
        try:
            grid_file_path = os.path.join(self.basepath,self.meta[self.name]['grid'])
        except LookupError:
            warnings.warn('Could not find a reference to a grid file in the metadata.')
            grid_file_path = ''
        if buffered:
            self.data = np.load(self.path)
        else:
            self.data = np.load(self.path,mmap_mode='c')
        self.small_pos = small_pos
        self.mesh = get_mesh_pts(grid_file_path,self.data.shape)
        self.data *= self.axes[4][self.units+' multiplier']
        for a,mesh_pts in enumerate(self.mesh):
            mesh_pts *= self.axes[a][self.units+' multiplier']

    def name(self):
        return self.name
    def get(self):
        return self.data
    def set(self,new_data):
        self.data = new_data
    def scale(self,factor):
        self.data *= factor
    def max_frame(self):
        return self.data.shape[0]-1
    def phys_dims4(self):
        if self.mesh==None:
            span = [0.0,1.0]*4
        else:
            span = []
            for ax in range(4):
                m = self.mesh[ax]
                if len(m)<2:
                    span += [0.0,1.0]
                else:
                    span += [1.5*m[0]-0.5*m[1] , 1.5*m[-1]-0.5*m[-2]]
        return np.array(span)
    def dims4(self):
        return self.data.shape
    def display_info(self):
        print('-------------------------')
        print('grid cells =',self.data.shape)
        print('physical size =',self.phys_dims4())
        print('-------------------------')
    def plot_coord(self,dict,ax,idx):
        # N.b. ax is relative, i.e., 0=hor., 1=ver.
        ds = (dict['extent'][2*ax+1] - dict['extent'][2*ax])/dict['dims'][ax]
        return dict['extent'][2*ax] + ds*idx + ds/2
    def get_global_minmax(self,dyn_range):
        if dyn_range==0.0:
            return np.min(self.data),np.max(self.data)
        else:
            logmax = np.log10(np.max(np.abs(self.data))+self.small_pos)
            return logmax-dyn_range , logmax
    def update_dyn_range(self,plot_dict,data_slice,dyn_range):
        lab = plot_dict['blabel']
        if dyn_range==0.0:
            plot_dict['vmin'] = np.min(data_slice)
            plot_dict['vmax'] = np.max(data_slice)
            if len(lab)>11:
                if lab[:11]=='log$_{10}$ ':
                    plot_dict['blabel'] = lab[11:]
        else:
            maxval = np.max(np.abs(data_slice))
            logmax = np.log10(maxval+self.small_pos)
            plot_dict['vmin'] = logmax-dyn_range
            plot_dict['vmax'] = logmax
            data_slice[...] = np.log10(np.abs(data_slice)+self.small_pos)
            if len(lab)>11:
                if lab[:11]!='log$_{10}$ ':
                    plot_dict['blabel'] = 'log$_{10}$ ' + lab
            else:
                plot_dict['blabel'] = 'log$_{10}$ ' + lab

    def start_plot(self,slicing_spec='zxty'):
        '''Not directly called.
        Returns an initial plotting dictionary for use with any kind of plot'''
        if len(slicing_spec)!=4:
            raise TypeError('The slicing_spec was not 4 characters.')
        axes = get_axis_info(slicing_spec)
        phys_dims = self.phys_dims4()
        plot_dict = {}

        lab = [self.axes[axes[i]]['label'] for i in range(5)]
        ulab = [self.axes[axes[i]][self.units+' label'] for i in range(5)]
        dkey = ['xlabel','ylabel','zlabel','none','blabel']
        if self.units=='plasma':
            for i in [0,1,2,4]:
                if ulab[i]=='None':
                    plot_dict[dkey[i]] = lab[i]
                else:
                    s = ulab[i].find('/')
                    if s==-1:
                        plot_dict[dkey[i]] = ulab[i] + lab[i]
                    else:
                        plot_dict[dkey[i]] = (ulab[i][:s] + ' $ ' + lab[i] + ' $ ' + ulab[i][s:]).replace('$ $','')
        else:
            for i in [0,1,2,4]:
                if ulab[i]=='None':
                    plot_dict[dkey[i]] = lab[i]
                else:
                    plot_dict[dkey[i]] = lab[i] + ' (' + ulab[i] + ')'
        plot_dict['dims'] = [self.dims4()[axes[i]] for i in range(3)]
        plot_dict['xpts'] = self.mesh[axes[0]]
        plot_dict['ypts'] = self.mesh[axes[1]]
        plot_dict['zpts'] = self.mesh[axes[2]]
        return plot_dict

    def volume3d(self,slicing_spec='xyzt',slice_to_plot=0,dyn_range=0.0):
        '''Returns data,dict
        data = the numpy array with the 3D slice data
        dict = a plotting dictionary to be used externally with mayavi
        The dictionary keys often have the same names as mayavi function arguments.  Available keys:
        "extent", "xlabel", "ylabel", "zlabel", "blabel", "dims", "vmin", "vmax", "mesh"
        Inputs:
        slicing_spec is a 4 character code ordering the axes, e.g., zxty
        The first 3 characters are the plotting volume, the last is the slicing axis
        slice_to_plot is an integer selecting a point on the slicing axis'''
        if len(slicing_spec)!=4:
            raise TypeError('The slicing_spec was not 4 characters.')
        if not slicing_spec=='xyzt':
            raise ValueError('The plotting axes have to be in order at present.')
        axes = get_axis_info(slicing_spec)
        phys_dims = self.phys_dims4()
        plot_dict = self.start_plot(slicing_spec)
        data_slice = np.take(self.data,[slice_to_plot],axis=axes[3])
        data_slice = np.squeeze(data_slice,axis=(axes[3],))

        ext = [[phys_dims[axes[i]*2],phys_dims[axes[i]*2+1]] for i in range(3)]
        plot_dict['extent'] = ext[0]+ext[1]+ext[2]
        self.update_dyn_range(plot_dict,data_slice,dyn_range)
        return data_slice,plot_dict

    def falsecolor2d(self,slicing_spec='zxty',slice_to_plot=(0,0),dyn_range=0.0,lib='matplotlib'):
        '''Returns data,dict
        data = the numpy array with the 2D slice data
        dict = a plotting dictionary to be used externally with MPL
        The dictionary keys have the same names as MPL function arguments.  Available keys:
        "extent", "xlabel", "ylabel", "dims", "aspect", "vmin", "vmax"
        Inputs:
        slicing_spec is a 4 character code ordering the axes, e.g., zxty
        The first two characters are the plotting plane, the next two are the slicing plane
        slice_to_plot is a tuple (s3,s4) selecting a point in the slicing plane'''
        if len(slice_to_plot)!=2:
            raise TypeError('The slice for falsecolor2d is supposed to be a 2-tuple.')
        #my_aspect = 'equal'
        my_aspect = 'auto'
        axes = get_axis_info(slicing_spec)
        phys_dims = self.phys_dims4()
        plot_dict = self.start_plot(slicing_spec)
        data_slice = np.take(self.data,[slice_to_plot[0]],axis=axes[2])
        data_slice = np.take(data_slice,[slice_to_plot[1]],axis=axes[3])
        data_slice = np.squeeze(data_slice,axis=(axes[2],axes[3]))

        # matplotlib plot data is accessed as [v,h]
        # mayavi plot data is accessed as [h,v]
        # twutils is supposed to put horizontal as the axis listed first
        if axes[0]<axes[1] and lib=='matplotlib':
            data_slice = data_slice.swapaxes(0,1)
        if axes[1]<axes[0] and lib=='mayavi':
            data_slice = data_slice.swapaxes(0,1)

        ext = [[phys_dims[axes[i]*2],phys_dims[axes[i]*2+1]] for i in range(2)]
        plot_dict['extent'] = ext[0]+ext[1]
        plot_dict['aspect'] = my_aspect
        self.update_dyn_range(plot_dict,data_slice,dyn_range)
        return data_slice,plot_dict

    def lineout(self,slicing_spec='xyzt',slice_to_plot=(0,0,0),dyn_range=0.0):
        '''Returns abcissa,ordinate,dict
        abcissa,ordinate = the numpy arrays with the 1D axis values and data
        dict = a plotting dictionary to be used externally with MPL
        The dictionary keys have the same names as MPL function arguments. Available keys:
        "extent", "xlabel", "ylabel", "dims"
        Inputs:
        slicing_spec is a 4 character code ordering the axes, e.g., zxty
        The first character is the plotting axis, the next three are the slicing volume
        slice_to_plot is a tuple (s2,s3,s4) selecting a point in the slicing volume'''
        if len(slice_to_plot)!=3:
            raise TypeError('The slice for lineout is supposed to be a 3-tuple.')
        axes = get_axis_info(slicing_spec)
        phys_dims = self.phys_dims4()
        plot_dict = self.start_plot(slicing_spec)
        data_slice = np.take(self.data,[slice_to_plot[0]],axis=axes[1])
        data_slice = np.take(data_slice,[slice_to_plot[1]],axis=axes[2])
        data_slice = np.take(data_slice,[slice_to_plot[2]],axis=axes[3])
        data_slice = np.squeeze(data_slice,axis=(axes[1],axes[2],axes[3]))
        if self.mesh!=None:
            abcissa = self.mesh[axes[0]]
        else:
            abcissa = np.linspace(0.0,1.0,len(data_slice))
        self.update_dyn_range(plot_dict,data_slice,dyn_range)
        plot_dict['ylabel'] = plot_dict['blabel']
        return abcissa,data_slice,plot_dict



'''
Module :samp:`twutils.pre`
--------------------------

This module contains functions and classes to help with setting parameters for turboWAVE input files.'''

import numpy as np
import scipy.constants as C

class SimUnits:
	"""Contains constants characterizing plasma units.
	Class is created with the unit density in mks units.

	t1 = unit of time in seconds

	x1 = unit of length in meters

	E1 = unit of electric field in volts per meter"""
	t1 = 0.0
	x1 = 0.0
	E1 = 0.0
	def __init__(self,n1):
		wp = np.sqrt(n1*C.e*C.e/(C.epsilon_0*C.m_e))
		self.t1 = 1/wp
		self.x1 = C.c/wp
		self.E1 = C.m_e * C.c * wp / C.e

class QuinticPulseData:
	"""Contains constants characterizing a quintic pulse with unit amplitude and base-to-base duration.

	energy = integration over square of quintic pulse

	FWHM = FWHM of square of quintic pulse"""
	energy = 0.3918
	FWHM = 0.3856

def ncrit(wavelength):
	"""Returns critical plasma density given a wavelength (mks units)"""
	w0 = 2*np.pi*C.c/wavelength
	return w0**2 * C.epsilon_0 * C.m_e / C.e**2

def courant(dx,dy,dz,c=1.0):
	"""Returns maximum time step allowed by the vacuum Courant condition.

	:param float dx: cell size in x
	:param float dy: cell size in y
	:param float dz: cell size in z
	:param float c: speed of light in the preferred units (defaults to plasma units)

	:return: maximum time step"""
	return 1/np.sqrt(1/dx/dx + 1/dy/dy + 1/dz/dz)/c

def getexp(a0,w0,r0,risetime,n1):
	"""Get experimental pulse parameters from simulation parameters.
	Inputs are in plasma units, outputs in mks.

	:param float a0: vector potential
	:param float w0: angular frequency
	:param float r0: 1/e of amplitude radius
	:param float risetime: quintic rise parameter
	:param float n1: unit density in *mks* units

	:return: (intensity,power,energy,FWHM)"""
	su = SimUnits(n1)
	qpd = QuinticPulseData()
	eta0 = np.sqrt(C.mu_0/C.epsilon_0)
	rmks = r0*su.x1
	base_to_base = 2.0*risetime*su.t1
	A0 = C.m_e * C.c * a0 / C.e
	wp = np.sqrt(n1*C.e*C.e/(C.epsilon_0*C.m_e))
	E0 = (w0/su.t1) * A0
	I0 = 0.5*E0*E0/eta0
	P0 = 0.5*np.pi*rmks*rmks*I0
	U0 = P0 * qpd.energy * base_to_base
	FWHM = qpd.FWHM * base_to_base
	return I0,P0,U0,FWHM

def getsim(U0,l0,r0,FWHM,n1):
	"""Get simulation pulse parameters from experimental parameters.
	Inputs are in mks, outputs in plasma units.

	:param float U0: pulse energy
	:param float l0: central wavelength
	:param float r0: 1/e of amplitude radius
	:param float FWHM: pulse length, full-width-half-maximum of intensity
	:param float n1: unit density

	:return: (vector potential,angular frequency,quintic risetime)"""
	su = SimUnits(n1)
	qpd = QuinticPulseData()
	eta0 = np.sqrt(C.mu_0/C.epsilon_0)
	w0_sim = su.t1 * 2.0*np.pi*C.c / l0
	r_sim = r0/su.x1
	base_to_base = FWHM / qpd.FWHM
	P0 = U0 / (qpd.energy * base_to_base)
	I0 = 2.0*P0 / (np.pi * r0 * r0)
	E0 = np.sqrt(2.0*I0*eta0)
	A0 = su.t1 * E0 / w0_sim
	a0 = C.e * A0 / (C.m_e * C.c)
	risetime = 0.5 * base_to_base / su.t1
	return a0,w0_sim,risetime
